# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

from io import StringIO
from pyomo.common.log import LoggingIntercept

import pyomo.common.unittest as unittest
import random

from pyomo.opt import check_available_solvers
from pyomo.environ import (
    ConcreteModel,
    Set,
    Objective,
    Constraint,
    Var,
    Block,
    Param,
    NonNegativeReals,
    TransformationFactory,
    ComponentUID,
    inequality,
)

import pyomo.core.expr as EXPR
from pyomo.core.expr.compare import assertExpressionsEqual

solvers = check_available_solvers('glpk')


class TestAddSlacks(unittest.TestCase):
    def setUp(self):
        # set seed so we can test name collisions predictably
        random.seed(666)

    @staticmethod
    def makeModel():
        model = ConcreteModel()
        model.x = Var(within=NonNegativeReals)
        model.y = Var(within=NonNegativeReals)
        model.rule1 = Constraint(expr=model.x <= 5)
        model.rule2 = Constraint(expr=inequality(1, model.y, 3))
        model.rule3 = Constraint(expr=model.x >= 0.1)
        model.obj = Objective(expr=-model.x - model.y)
        return model

    def test_add_trans_block(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m)

        xblock = m.component("_core_add_slack_variables")
        self.assertIsInstance(xblock, Block)

    def test_trans_block_name_collision(self):
        m = self.makeModel()
        m._core_add_slack_variables = Block()
        TransformationFactory('core.add_slack_variables').apply_to(m)
        xblock = m.component("_core_add_slack_variables_4")
        self.assertIsInstance(xblock, Block)

    def test_slack_vars_added(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m)
        xblock = m.component("_core_add_slack_variables")

        # should have new variables on new block
        self.assertIsInstance(xblock.component("_slack_minus_rule1"), Var)
        self.assertFalse(hasattr(xblock, "_slack_plus_rule1"))
        self.assertIsInstance(xblock.component("_slack_minus_rule2"), Var)
        self.assertIsInstance(xblock.component("_slack_plus_rule2"), Var)
        self.assertFalse(hasattr(xblock, "_slack_minus_rule3"))
        self.assertIsInstance(xblock.component("_slack_plus_rule3"), Var)
        # all new variables in non-negative reals
        self.assertEqual(xblock._slack_minus_rule1.bounds, (0, None))
        self.assertEqual(xblock._slack_minus_rule2.bounds, (0, None))
        self.assertEqual(xblock._slack_plus_rule2.bounds, (0, None))
        self.assertEqual(xblock._slack_plus_rule3.bounds, (0, None))

    # wrapping this as a method because I use it again later when I test
    # targets
    def checkRule1(self, m):
        # check all original variables still there:
        cons = m.rule1
        transBlock = m.component("_core_add_slack_variables")

        self.assertIsNone(cons.lower)
        self.assertEqual(cons.upper, 5)
        assertExpressionsEqual(
            self,
            cons.body,
            EXPR.LinearExpression(
                [m.x, EXPR.MonomialTermExpression((-1, transBlock._slack_minus_rule1))]
            ),
        )

    def checkRule3(self, m):
        # check all original variables still there:
        cons = m.rule3
        transBlock = m.component("_core_add_slack_variables")

        self.assertIsNone(cons.upper)
        self.assertEqual(cons.lower, 0.1)

        assertExpressionsEqual(
            self, cons.body, EXPR.LinearExpression([m.x, transBlock._slack_plus_rule3])
        )

    def test_ub_constraint_modified(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m)
        self.checkRule1(m)

    def test_lb_constraint_modified(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m)
        self.checkRule3(m)

    def test_both_bounds_constraint_modified(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m)

        # check all original variables still there:
        cons = m.rule2
        transBlock = m.component("_core_add_slack_variables")

        self.assertEqual(cons.lower, 1)
        self.assertEqual(cons.upper, 3)

        assertExpressionsEqual(
            self,
            cons.body,
            EXPR.LinearExpression(
                [
                    m.y,
                    transBlock._slack_plus_rule2,
                    EXPR.MonomialTermExpression((-1, transBlock._slack_minus_rule2)),
                ]
            ),
        )

    def test_obj_deactivated(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m)

        # should have old objective deactivated.
        self.assertFalse(m.obj.active)

    def test_new_obj_created(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m)

        transBlock = m.component("_core_add_slack_variables")

        # active objective should minimize sum of slacks
        obj = transBlock.component("_slack_objective")
        self.assertIsInstance(obj, Objective)
        self.assertTrue(obj.active)

        assertExpressionsEqual(
            self,
            obj.expr,
            EXPR.LinearExpression(
                [
                    transBlock._slack_minus_rule1,
                    transBlock._slack_plus_rule2,
                    transBlock._slack_minus_rule2,
                    transBlock._slack_plus_rule3,
                ]
            ),
        )

    def test_badModel_err(self):
        model = ConcreteModel()
        model.x = Var(within=NonNegativeReals)
        model.rule1 = Constraint(expr=inequality(6, model.x, 5))
        self.assertRaisesRegex(
            RuntimeError,
            "Lower bound exceeds upper bound in constraint rule1*",
            TransformationFactory('core.add_slack_variables').apply_to,
            model,
        )

    def test_leave_deactivated_constraints(self):
        m = self.makeModel()
        m.rule2.deactivate()
        TransformationFactory('core.add_slack_variables').apply_to(m)

        cons = m.rule2
        self.assertFalse(cons.active)
        self.assertEqual(cons.lower, 1)
        self.assertEqual(cons.upper, 3)
        # cons.body is a ScalarVar
        self.assertIs(cons.body, m.y)

    def checkTargetSlackVar(self, transBlock):
        self.assertIsNone(transBlock.component("_slack_minus_rule2"))
        self.assertIsNone(transBlock.component("_slack_plus_rule2"))
        self.assertFalse(hasattr(transBlock, "_slack_minus_rule3"))
        self.assertIsInstance(transBlock.component("_slack_plus_rule3"), Var)

    def checkTargetSlackVars(self, transBlock):
        self.assertIsInstance(transBlock.component("_slack_minus_rule1"), Var)
        self.assertFalse(hasattr(transBlock, "_slack_plus_rule1"))
        self.assertIsNone(transBlock.component("_slack_minus_rule2"))
        self.assertIsNone(transBlock.component("_slack_plus_rule2"))
        self.assertFalse(hasattr(transBlock, "_slack_minus_rule3"))
        self.assertIsInstance(transBlock.component("_slack_plus_rule3"), Var)

    def test_only_targets_have_slack_vars(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(
            m, targets=[m.rule1, m.rule3]
        )

        transBlock = m.component("_core_add_slack_variables")
        # check that we only made slack vars for targets
        self.checkTargetSlackVars(transBlock)

    def test_only_targets_have_slack_vars_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1, m.rule3]
        )

        transBlock = m2.component("_core_add_slack_variables")
        # check that we only made slack vars for targets
        self.checkTargetSlackVars(transBlock)

    def checkNonTargetCons(self, m):
        cons = m.rule2
        self.assertEqual(cons.lower, 1)
        self.assertEqual(cons.upper, 3)
        # cons.body is a ScalarVar
        self.assertIs(cons.body, m.y)

    def test_nontarget_constraint_same(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(
            m, targets=[m.rule1, m.rule3]
        )

        self.checkNonTargetCons(m)

    def test_nontarget_constraint_same_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1, m.rule3]
        )

        self.checkNonTargetCons(m2)

    def test_target_constraints_transformed(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(
            m, targets=[m.rule1, m.rule3]
        )

        self.checkRule1(m)
        self.checkRule3(m)

    def test_target_constraints_transformed_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1, m.rule3]
        )

        self.checkRule1(m2)
        self.checkRule3(m2)

    def checkTargetObj(self, m):
        transBlock = m._core_add_slack_variables
        obj = transBlock.component("_slack_objective")
        self.assertIs(obj.expr, transBlock._slack_plus_rule3)

    def checkTargetsObj(self, m):
        transBlock = m._core_add_slack_variables
        obj = transBlock.component("_slack_objective")
        assertExpressionsEqual(
            self,
            obj.expr,
            EXPR.LinearExpression(
                [transBlock._slack_minus_rule1, transBlock._slack_plus_rule3]
            ),
        )

    def test_target_objective(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(
            m, targets=[m.rule1, m.rule3]
        )

        self.assertFalse(m.obj.active)
        self.checkTargetsObj(m)

    def test_target_objective_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1, m.rule3]
        )

        self.assertFalse(m2.obj.active)
        self.checkTargetsObj(m2)

    def test_err_for_bogus_kwds(self):
        m = self.makeModel()
        self.assertRaisesRegex(
            ValueError,
            "key 'notakwd' not defined for ConfigDict ''",
            TransformationFactory('core.add_slack_variables').apply_to,
            m,
            notakwd="I want a feasible model",
        )

    def test_error_for_non_constraint_noniterable_target(self):
        m = self.makeModel()
        m.indexedVar = Var([1, 2])
        self.assertRaisesRegex(
            ValueError,
            "Expected Constraint or list of Constraints.\n\tReceived "
            "<class 'pyomo.core.base.var.VarData'>",
            TransformationFactory('core.add_slack_variables').apply_to,
            m,
            targets=m.indexedVar[1],
        )

    def test_error_for_non_constraint_target_in_list(self):
        m = self.makeModel()
        self.assertRaisesRegex(
            ValueError,
            "Expected Constraint or list of Constraints.\n\tReceived "
            "<class 'pyomo.core.base.var.ScalarVar'>",
            TransformationFactory('core.add_slack_variables').apply_to,
            m,
            targets=[m.rule1, m.x],
        )

    def test_deprecation_warning_for_cuid_target(self):
        m = self.makeModel()
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.core'):
            TransformationFactory('core.add_slack_variables').apply_to(
                m, targets=ComponentUID(m.rule3)
            )
        self.assertRegex(
            out.getvalue(),
            "DEPRECATED: In future releases ComponentUID "
            "targets will no longer be\nsupported in the "
            "core.add_slack_variables transformation. "
            "Specify\ntargets as a Constraint or list of "
            "Constraints.*",
        )

        # make sure that it still worked though
        self.checkNonTargetCons(m)
        self.checkRule3(m)
        self.assertFalse(m.obj.active)
        self.checkTargetObj(m)
        transBlock = m.component("_core_add_slack_variables")
        self.checkTargetSlackVar(transBlock)

    def test_deprecation_warning_for_cuid_targets(self):
        m = self.makeModel()
        out = StringIO()
        with LoggingIntercept(out, 'pyomo.core'):
            TransformationFactory('core.add_slack_variables').apply_to(
                m, targets=[ComponentUID(m.rule1), ComponentUID(m.rule3)]
            )
        self.assertRegex(
            out.getvalue(),
            "DEPRECATED: In future releases ComponentUID "
            "targets will no longer be\nsupported in the "
            "core.add_slack_variables transformation. "
            "Specify\ntargets as a Constraint or list of "
            "Constraints.*",
        )
        # make sure that it still worked though
        self.checkNonTargetCons(m)
        self.checkRule1(m)
        self.checkRule3(m)
        self.assertFalse(m.obj.active)
        self.checkTargetsObj(m)
        transBlock = m.component("_core_add_slack_variables")
        self.checkTargetSlackVars(transBlock)

    def test_transformed_constraints_sumexpression_body(self):
        m = self.makeModel()
        m.rule4 = Constraint(expr=inequality(5, m.x - 2 * m.y, 9))
        TransformationFactory('core.add_slack_variables').apply_to(m, targets=m.rule4)

        transBlock = m._core_add_slack_variables
        c = m.rule4
        self.assertEqual(c.lower, 5)
        self.assertEqual(c.upper, 9)

        assertExpressionsEqual(
            self,
            c.body,
            EXPR.LinearExpression(
                [
                    m.x,
                    EXPR.MonomialTermExpression((-2, m.y)),
                    transBlock._slack_plus_rule4,
                    EXPR.MonomialTermExpression((-1, transBlock._slack_minus_rule4)),
                ]
            ),
        )

    def test_transformed_constraint_scalar_body(self):
        m = self.makeModel()
        m.p = Param(initialize=6, mutable=True)
        m.rule4 = Constraint(expr=m.p <= 9)
        TransformationFactory('core.add_slack_variables').apply_to(m, targets=[m.rule4])

        transBlock = m._core_add_slack_variables
        c = m.rule4
        self.assertIsNone(c.lower)
        self.assertEqual(c.upper, 9)
        self.assertEqual(c.body.nargs(), 2)
        self.assertEqual(c.body.arg(0).value, 6)
        self.assertIs(c.body.arg(1).__class__, EXPR.MonomialTermExpression)
        self.assertEqual(c.body.arg(1).arg(0), -1)
        self.assertIs(c.body.arg(1).arg(1), transBlock._slack_minus_rule4)


class TestAddSlacks_IndexedConstraints(unittest.TestCase):
    @staticmethod
    def makeModel():
        m = ConcreteModel()
        m.S = Set(initialize=[1, 2, 3])
        m.x = Var(m.S)
        m.y = Var()

        def rule1_rule(m, s):
            return 2 * m.x[s] >= 4

        m.rule1 = Constraint(m.S, rule=rule1_rule)
        m.rule2 = Constraint(expr=m.y <= 6)
        m.obj = Objective(expr=sum(m.x[s] for s in m.S) - m.y)
        return m

    def checkSlackVars_indexedtarget(self, transBlock):
        self.assertIsInstance(transBlock.component("_slack_plus_rule1[1]"), Var)
        self.assertIsInstance(transBlock.component("_slack_plus_rule1[2]"), Var)
        self.assertIsInstance(transBlock.component("_slack_plus_rule1[3]"), Var)
        self.assertIsNone(transBlock.component("_slack_minus_rule2"))

    def test_indexedtarget_only_create_slackvars_for_targets(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m, targets=[m.rule1])

        transBlock = m.component("_core_add_slack_variables")
        # TODO: So, right now indexed constraints don't result in indexed
        # slack variables. They could... But I don't know if it matters much?
        # They are named sensibly either way... Dunno.
        self.checkSlackVars_indexedtarget(transBlock)

    def test_indexedtarget_only_create_slackvars_for_targets_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1]
        )

        transBlock = m2.component("_core_add_slack_variables")
        self.checkSlackVars_indexedtarget(transBlock)

    def checkRule2(self, m):
        cons = m.rule2
        self.assertEqual(cons.upper, 6)
        self.assertIsNone(cons.lower)
        self.assertIs(cons.body, m.y)

    def test_indexedtarget_nontarget_same(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m, targets=[m.rule1])

        self.checkRule2(m)

    def test_indexedtarget_nontarget_same_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1]
        )

        self.checkRule2(m2)

    def checkTargetObj(self, m):
        transBlock = m._core_add_slack_variables
        obj = transBlock.component("_slack_objective")
        self.assertIsInstance(obj, Objective)
        assertExpressionsEqual(
            self,
            obj.expr,
            EXPR.LinearExpression(
                [
                    transBlock.component("_slack_plus_rule1[1]"),
                    transBlock.component("_slack_plus_rule1[2]"),
                    transBlock.component("_slack_plus_rule1[3]"),
                ]
            ),
        )

    def test_indexedtarget_objective(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m, targets=[m.rule1])

        self.assertFalse(m.obj.active)
        self.checkTargetObj(m)

    def test_indexedtarget_objective_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1]
        )

        self.assertFalse(m2.obj.active)
        self.checkTargetObj(m2)

    def checkTransformedRule1(self, m, i):
        c = m.rule1[i]
        self.assertEqual(c.lower, 4)
        self.assertIsNone(c.upper)

        assertExpressionsEqual(
            self,
            c.body,
            EXPR.LinearExpression(
                [
                    EXPR.MonomialTermExpression((2, m.x[i])),
                    m._core_add_slack_variables.component("_slack_plus_rule1[%s]" % i),
                ]
            ),
        )

    def test_indexedtarget_targets_transformed(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(m, targets=[m.rule1])

        for i in [1, 2, 3]:
            self.checkTransformedRule1(m, i)

    def test_indexedtarget_targets_transformed_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=m.rule1
        )

        for i in [1, 2, 3]:
            self.checkTransformedRule1(m2, i)

    def checkSlackVars_constraintDataTarget(self, transBlock):
        self.assertIsInstance(transBlock.component("_slack_plus_rule1[2]"), Var)
        self.assertIsNone(transBlock.component("_slack_plus_rule1[1]"))
        self.assertIsNone(transBlock.component("_slack_plus_rule1[3]"))
        self.assertIsNone(transBlock.component("_slack_minus_rule2"))

    def test_ConstraintDatatarget_only_add_slackvars_for_targets(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(
            m, targets=[m.rule1[2]]
        )

        transBlock = m._core_add_slack_variables
        self.checkSlackVars_constraintDataTarget(transBlock)

    def test_ConstraintDatatarget_only_add_slackvars_for_targets_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=m.rule1[2]
        )

        transBlock = m2._core_add_slack_variables
        self.checkSlackVars_constraintDataTarget(transBlock)

    def checkUntransformedRule1(self, m, i):
        c = m.rule1[i]
        self.assertEqual(c.lower, 4)
        self.assertIsNone(c.upper)
        self.assertEqual(c.body.arg(0), 2)
        self.assertIs(c.body.arg(1), m.x[i])

    def test_ConstraintDatatarget_nontargets_same(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(
            m, targets=[m.rule1[2]]
        )

        self.checkUntransformedRule1(m, 1)
        self.checkUntransformedRule1(m, 3)
        self.checkRule2(m)

    def test_ConstraintDatatarget_nontargets_same_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1[2]]
        )

        self.checkUntransformedRule1(m2, 1)
        self.checkUntransformedRule1(m2, 3)
        self.checkRule2(m2)

    def test_ConstraintDatatarget_target_transformed(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(
            m, targets=[m.rule1[2]]
        )

        self.checkTransformedRule1(m, 2)

    def test_ConstraintDatatarget_target_transformed_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1[2]]
        )

        self.checkTransformedRule1(m2, 2)

    def checkConstraintDataObj(self, m):
        transBlock = m._core_add_slack_variables
        obj = transBlock.component("_slack_objective")
        self.assertIsInstance(obj, Objective)
        self.assertIs(obj.expr, transBlock.component("_slack_plus_rule1[2]"))

    def test_ConstraintDatatarget_objective(self):
        m = self.makeModel()
        TransformationFactory('core.add_slack_variables').apply_to(
            m, targets=[m.rule1[2]]
        )

        self.assertFalse(m.obj.active)
        self.checkConstraintDataObj(m)

    def test_ConstraintDatatarget_objective_create_using(self):
        m = self.makeModel()
        m2 = TransformationFactory('core.add_slack_variables').create_using(
            m, targets=[m.rule1[2]]
        )

        self.assertFalse(m2.obj.active)
        self.checkConstraintDataObj(m2)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
UI Tests
"""

# The pytest-qt plugin can generate exceptions / core dumps when it is
# run in a terminal (without an active X11 screen).  Setting the
# QT_QPA_PLATFORM environment variable *before* initializing Qt can work
# around this error (see https://stackoverflow.com/a/74719383):
import os

os.environ['QT_QPA_PLATFORM'] = 'offscreen'

from pyomo.environ import (
    ConcreteModel,
    Var,
    BooleanVar,
    Param,
    Constraint,
    Objective,
    Reals,
    Block,
    Expression,
    ExternalFunction,
    sin,
    sqrt,
    log,
    value,
)
import pyomo.common.unittest as unittest
import pyomo.contrib.viewer.qt as myqt
import pyomo.contrib.viewer.pyomo_viewer as pv
from pyomo.contrib.viewer.qt import available
from pyomo.core.base.units_container import pint_available

if available:
    import contextvars
    from pyomo.contrib.viewer.qt import QtCore, QMessageBox
    from pyomo.contrib.viewer.ui import get_mainwindow, ModelBrowser
else:
    # If qt is not available, we still need to have a fake pytest fixture
    # in order to stop errors
    @unittest.pytest.fixture(scope="module")
    def qtbot():
        """Overwrite qtbot - remove test failure"""
        return

    pytestmark = unittest.pytest.mark.skip("Qt components are not available.")

if not pint_available:
    pytestmark = unittest.pytest.mark.skip(
        "contrib.viewer requires pint, which is not available."
    )

if not pv.qtconsole_available:
    pytestmark = unittest.pytest.mark.skip(
        "contrib.viewer requires qtconsole, which is not available."
    )


def get_model():
    # Borrowed this test model from the trust region tests
    m = ConcreteModel()
    m.y = BooleanVar(range(3), initialize=False)
    m.z = Var(range(3), domain=Reals, initialize=2.0)
    m.x = Var(range(4), initialize=2.0)
    m.x[1] = 1.0
    m.x[2] = 0.0
    m.x[3] = None

    m.b1 = Block()
    m.b1.e1 = Expression(expr=m.x[0] + m.x[1])
    m.b1.e2 = Expression(expr=m.x[0] / m.x[2])
    m.b1.e3 = Expression(expr=m.x[3] * m.x[1])
    m.b1.e4 = Expression(expr=log(m.x[2]))
    m.b1.e5 = Expression(expr=log(m.x[2] - 2))

    def blackbox(a, b):
        return sin(a - b)

    m.bb = ExternalFunction(blackbox)

    m.obj = Objective(
        expr=(m.z[0] - 1.0) ** 2
        + (m.z[0] - m.z[1]) ** 2
        + (m.z[2] - 1.0) ** 2
        + (m.x[0] - 1.0) ** 4
        + (m.x[1] - 1.0) ** 6  # + m.bb(m.x[0],m.x[1])
    )
    m.c1 = Constraint(expr=m.x[0] * m.z[0] ** 2 + m.bb(m.x[0], m.x[1]) == 2 * sqrt(2.0))
    m.c2 = Constraint(expr=m.z[2] ** 4 * m.z[1] ** 2 + m.z[1] == 8 + sqrt(2.0))
    m.c3 = Constraint(expr=m.x[1] == 3)
    m.c4 = Constraint(expr=0 == 3 / m.x[2])
    m.c5 = Constraint(expr=0 == log(m.x[2]))
    m.c6 = Constraint(expr=0 == log(m.x[2] - 4))
    m.c7 = Constraint(expr=0 == log(m.x[3]))
    m.p1 = Param(mutable=True, initialize=1)
    m.c8 = Constraint(expr=m.x[1] <= 1 / m.p1)
    m.p1 = 0
    return m


def test_get_mainwindow(qtbot):
    m = get_model()
    mw = get_mainwindow(model=m, testing=True)
    assert hasattr(mw, "menuBar")
    assert isinstance(mw.variables, ModelBrowser)
    assert isinstance(mw.constraints, ModelBrowser)
    assert isinstance(mw.expressions, ModelBrowser)
    assert isinstance(mw.parameters, ModelBrowser)


def test_close_mainwindow(qtbot):
    mw = get_mainwindow(model=None, testing=True)
    mw.exit_action()


def test_show_model_select_no_models(qtbot):
    mw = get_mainwindow(model=None, testing=True)
    ms = mw.show_model_select()
    ms.update_models()
    ms.select_model()


def test_model_information(qtbot):
    m = get_model()
    mw = get_mainwindow(model=m, testing=True)
    mw.model_information()
    assert isinstance(mw._dialog, QMessageBox)
    text = mw._dialog.text()
    mw._dialog.close()
    text = text.split("\n")
    assert str(text[0]).startswith("8")  # Active constraints
    assert str(text[1]).startswith("7")  # Active equalities
    assert str(text[2]).startswith("7")  # Free vars in active equalities
    assert str(text[3]).startswith("0")  # degrees of feedom
    # Main window has parts it is supposed to
    assert hasattr(mw, "menuBar")
    assert isinstance(mw.variables, ModelBrowser)
    assert isinstance(mw.constraints, ModelBrowser)
    assert isinstance(mw.expressions, ModelBrowser)
    assert isinstance(mw.parameters, ModelBrowser)


def test_tree_expand_collapse(qtbot):
    m = get_model()
    mw = get_mainwindow(model=m, testing=True)
    mw.variables.treeView.expandAll()
    mw.variables.treeView.collapseAll()


def test_residual_table(qtbot):
    m = get_model()
    mw = get_mainwindow(model=m, testing=True)
    mw.residuals_restart()
    mw.ui_data.calculate_expressions()
    mw.residuals.calculate()
    mw.residuals_restart()
    mw.residuals.sort()
    dm = mw.residuals.tableView.model()
    # Name
    assert dm.data(dm.index(0, 0)) == "c4"
    # residual value
    assert dm.data(dm.index(0, 1)) == "Divide_by_0"
    # body value
    assert dm.data(dm.index(0, 2)) == "Divide_by_0"
    # upper
    assert dm.data(dm.index(0, 3)) == 0
    # lower
    assert dm.data(dm.index(0, 4)) == 0
    # active
    assert dm.data(dm.index(0, 5)) == True
    m.c4.deactivate()
    mw.residuals.sort()
    assert dm.data(dm.index(0, 0)) == "c5"


def test_var_tree(qtbot):
    m = get_model()
    mw = get_mainwindow(model=m, testing=True)
    qtbot.addWidget(mw)
    mw.variables.treeView.expandAll()
    root_index = mw.variables.datmodel.index(0, 0)
    z_index = mw.variables.datmodel.index(1, 0, parent=root_index)
    z_val_index = mw.variables.datmodel.index(1, 1, parent=root_index)
    z1_val_index = mw.variables.datmodel.index(0, 1, parent=z_index)
    assert mw.variables.datmodel.data(z1_val_index) == 2.0
    mw.variables.treeView.setCurrentIndex(z1_val_index)
    mw.variables.treeView.openPersistentEditor(z1_val_index)
    d = mw.variables.treeView.itemDelegate()
    w = mw.variables.treeView.indexWidget(z1_val_index)
    w.setText("Not a number")
    d.setModelData(w, mw.variables.datmodel, z1_val_index)
    assert (value(m.z[0]) - 2.0) < 1e-6  # unchanged
    w.setText("1e5")
    d.setModelData(w, mw.variables.datmodel, z1_val_index)
    assert (value(m.z[0]) - 1e5) < 1e-6  # set float
    w.setText("false")
    d.setModelData(w, mw.variables.datmodel, z1_val_index)
    assert (value(m.z[0]) - 0) < 1e-6  # bool to 0
    w.setText("true")
    d.setModelData(w, mw.variables.datmodel, z1_val_index)
    assert (value(m.z[0]) - 1) < 1e-6  # bool to 1
    mw.variables.treeView.closePersistentEditor(z1_val_index)
    w.setText("2")
    d.setModelData(w, mw.variables.datmodel, z1_val_index)
    assert (value(m.z[0]) - 2) < 1e-6  # set int
    mw.variables.treeView.closePersistentEditor(z1_val_index)


def test_bad_view(qtbot):
    m = get_model()
    mw = get_mainwindow(model=m, testing=True)
    err = None
    try:
        mw.badTree = mw._tree_restart(
            w=mw.variables, standard="Bad Stuff", ui_data=mw.ui_data
        )
    except ValueError:
        err = "ValueError"
    assert err == "ValueError"


def test_qtconsole_app(qtbot):
    app = pv.QtApp()
    # empty list to prevent picking up args from pytest
    app.initialize([])
    app.show_ui()
    app.hide_ui()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections
import functools
import itertools
import logging
import operator
import sys

from pyomo.common import enums
from pyomo.common.collections import Sequence, ComponentMap, ComponentSet
from pyomo.common.deprecation import deprecation_warning
from pyomo.common.errors import DeveloperError, InvalidValueError
from pyomo.common.numeric_types import (
    check_if_numeric_type,
    native_types,
    native_numeric_types,
    native_complex_types,
    native_logical_types,
)
from pyomo.core.pyomoobject import PyomoObject
from pyomo.core.base import (
    Var,
    Param,
    Expression,
    Objective,
    Block,
    Constraint,
    Expression,
    NumericLabeler,
    Suffix,
    SortComponents,
)
from pyomo.core.base.component import ActiveComponent
from pyomo.core.base.expression import NamedExpressionData
from pyomo.core.expr.numvalue import is_fixed, value
import pyomo.core.expr as EXPR
import pyomo.core.kernel as kernel

logger = logging.getLogger(__name__)

valid_expr_ctypes_minlp = {Var, Param, Expression, Objective}
valid_active_ctypes_minlp = {Block, Constraint, Objective, Suffix}
sum_like_expression_types = {
    EXPR.SumExpression,
    EXPR.LinearExpression,
    EXPR.NPV_SumExpression,
}
_named_subexpression_types = (
    NamedExpressionData,
    kernel.expression.expression,
    kernel.objective.objective,
)

HALT_ON_EVALUATION_ERROR = False
nan = float('nan')
int_float = {int, float}


class ExprType(enums.IntEnum):
    # Note that the ordering is meaningful, and we will compare
    # instances using relational operators.  In particular, we assume
    # that the Enum values increase in polynomial degree, starting at
    # CONSTANT and endine at GENERAL (nonlinear).
    CONSTANT = 0
    FIXED = 3
    VARIABLE = 5
    MONOMIAL = 10
    LINEAR = 20
    QUADRATIC = 30
    GENERAL = 40


_FileDeterminism_deprecation = {
    1: 20,
    2: 30,
    'DEPRECATED_KEYS': 20,
    'DEPRECATED_KEYS_AND_NAMES': 30,
}


class FileDeterminism(enums.IntEnum):
    NONE = 0
    # DEPRECATED_KEYS = 1
    # DEPRECATED_KEYS_AND_NAMES = 2
    ORDERED = 10
    SORT_INDICES = 20
    SORT_SYMBOLS = 30

    # We will define __str__ and __format__ so that behavior in python
    # 3.11+ is consistent with 3.7 - 3.10.

    def __str__(self):
        return enums.Enum.__str__(self)

    def __format__(self, spec):
        # Removal of Python 3.7 support allows us to use Enum.__format__
        return enums.Enum.__format__(self, spec)

    @classmethod
    def _missing_(cls, value):
        # This is not a perfect deprecation path, as the old attributes
        # are no longer valid.  However, as the previous implementation
        # was a pure int and not an Enum, this is sufficient for our
        # needs.
        if value in _FileDeterminism_deprecation:
            new = FileDeterminism(_FileDeterminism_deprecation[value])
            deprecation_warning(
                f'FileDeterminism({value}) is deprecated.  '
                f'Please use {str(new)} ({int(new)})',
                version='6.5.0',
            )
            return new
        return super()._missing_(value)


def val2str(val):
    """Converts an object to str with special handling for InvalidNumber

    This will convert the ``val`` to a string.  If ``val`` is an
    InvalidNumber, the conversion to string will bypass the exception
    raised by :py:meth:`InvalidNumber.__str__`.

    """
    if hasattr(val, '_str'):
        return val._str()
    if hasattr(val, 'to_string'):
        return val.to_string()
    return repr(val)


class InvalidNumber(PyomoObject):
    def __init__(self, value, cause=""):
        self.value = value
        if cause.__class__ is list:
            self.causes = list(cause)
        else:
            self.causes = [cause]

    @staticmethod
    def parse_args(*args):
        causes = []
        real_args = []
        for arg in args:
            if arg.__class__ is InvalidNumber:
                causes.extend(arg.causes)
                real_args.append(arg.value)
            else:
                real_args.append(arg)
        return real_args, causes

    def _cmp(self, op, other):
        args, causes = InvalidNumber.parse_args(self, other)
        try:
            return op(*args)
        except TypeError:
            # TypeError will be raised when comparing incompatible types
            # (e.g., int <= None)
            return False

    def _op(self, op, *args):
        args, causes = InvalidNumber.parse_args(*args)
        try:
            return InvalidNumber(op(*args), causes)
        except TypeError:
            # TypeError will be raised when operating on incompatible
            # types (e.g., int + None);
            return InvalidNumber(self.value, causes)
        except (ArithmeticError, ValueError):
            # ArithmeticError and ValueError can be raised by invalid
            # operations (like divide by zero or log of a negative number)
            return InvalidNumber(float('nan'), causes)

    def __eq__(self, other):
        ans = self._cmp(operator.eq, other)
        try:
            return bool(ans)
        except ValueError:
            # ValueError can be raised by numpy.ndarray when two arrays
            # are returned.  In that case, ndarray returns a new ndarray
            # of bool values.  We will fall back on using `all` to
            # reduce it to a single bool.
            try:
                return all(ans)
            except:
                pass
            raise

    def __lt__(self, other):
        # Note that as < is ambiguous for arrays, we will attempt to
        # cast the result to bool and if it was an array, allow the
        # exception to propagate
        return bool(self._cmp(operator.lt, other))

    def __gt__(self, other):
        # See the comment in __lt__() on the use of bool()
        return bool(self._cmp(operator.gt, other))

    def __le__(self, other):
        # See the comment in __lt__() on the use of bool()
        return bool(self._cmp(operator.le, other))

    def __ge__(self, other):
        # See the comment in __lt__() on the use of bool()
        return bool(self._cmp(operator.ge, other))

    def _error(self, msg):
        causes = list(filter(None, self.causes))
        if causes:
            msg += "\nThe InvalidNumber was generated by:\n\t"
            msg += "\n\t".join(causes)
        raise InvalidValueError(msg)

    def __str__(self):
        # We want attempts to convert InvalidNumber to a string
        # representation to raise a InvalidValueError, unless we are in
        # the middle of processing an exception.  In that case, it is
        # very likely that an exception handler is generating an error
        # message.  We will play nice and return a reasonable string.
        if sys.exc_info()[1] is None:
            self._error(f'Cannot emit {self._str()} in compiled representation')
        else:
            return self._str()

    def _str(self):
        return f'InvalidNumber({self.value!r})'

    def __repr__(self):
        return str(self)

    def __format__(self, format_spec):
        # FIXME: We want to move to where converting InvalidNumber to
        # string (with either repr() or f"") should raise a
        # InvalidValueError.  However, at the moment, this breaks some
        # tests in PyROS.
        # return self.value.__format__(format_spec)
        return self._error(f'Cannot emit {str(self)} in compiled representation')

    def __float__(self):
        # We want attempts to convert InvalidNumber to a float
        # representation to raise a InvalidValueError.
        return self._error(f'Cannot convert {str(self)} to float')

    def __neg__(self):
        return self._op(operator.neg, self)

    def __abs__(self):
        return self._op(operator.abs, self)

    def __add__(self, other):
        return self._op(operator.add, self, other)

    def __sub__(self, other):
        return self._op(operator.sub, self, other)

    def __mul__(self, other):
        return self._op(operator.mul, self, other)

    def __truediv__(self, other):
        return self._op(operator.truediv, self, other)

    def __pow__(self, other):
        return self._op(operator.pow, self, other)

    def __radd__(self, other):
        return self._op(operator.add, other, self)

    def __rsub__(self, other):
        return self._op(operator.sub, other, self)

    def __rmul__(self, other):
        return self._op(operator.mul, other, self)

    def __rtruediv__(self, other):
        return self._op(operator.truediv, other, self)

    def __rpow__(self, other):
        return self._op(operator.pow, other, self)


_CONSTANT = ExprType.CONSTANT


class BeforeChildDispatcher(collections.defaultdict):
    """Dispatcher for handling the :class:`StreamBasedExpressionVisitor`
    `beforeChild` callback

    This dispatcher implements a specialization of :class:`defaultdict`
    that supports automatic type registration.  Any missing types will
    return the :meth:`register_dispatcher` method, which (when called
    as a callback) will interrogate the type, identify the appropriate
    callback, add the callback to the dict, and return the result of
    calling the callback.  As the callback is added to the dict, no type
    will incur the overhead of `register_dispatcher` more than once.

    Note that all dispatchers are implemented as `staticmethod`
    functions to avoid the (unnecessary) overhead of binding to the
    dispatcher object.

    """

    __slots__ = ()

    def __missing__(self, key):
        return self.register_dispatcher

    def register_dispatcher(self, visitor, child):
        child_type = type(child)
        if child_type in native_numeric_types:
            self[child_type] = self._before_native_numeric
        elif child_type in native_logical_types:
            self[child_type] = self._before_native_logical
        elif issubclass(child_type, str):
            self[child_type] = self._before_string
        elif child_type in native_types:
            if issubclass(child_type, tuple(native_complex_types)):
                self[child_type] = self._before_complex
            else:
                self[child_type] = self._before_invalid
        elif not hasattr(child, 'is_expression_type'):
            if check_if_numeric_type(child):
                self[child_type] = self._before_native_numeric
            else:
                self[child_type] = self._before_invalid
        elif not child.is_expression_type():
            if child.is_indexed():
                cdata = child._ComponentDataClass(child)
                if cdata.is_expression_type():
                    self[child_type] = self._before_indexed_expr
                elif cdata.is_numeric_type() or child.is_logical_type():
                    if cdata.is_potentially_variable():
                        self[child_type] = self._before_indexed_var
                    else:
                        self[child_type] = self._before_indexed_param
                elif child.is_component_type():
                    self[child_type] = self._before_indexed_component
            elif child.is_numeric_type() or child.is_logical_type():
                if child.is_potentially_variable():
                    self[child_type] = self._before_var
                elif isinstance(child, EXPR.IndexTemplate):
                    self[child_type] = self._before_index_template
                else:
                    self[child_type] = self._before_param
            elif child.is_component_type():
                self[child_type] = self._before_component
        elif not child.is_potentially_variable():
            self[child_type] = self._before_npv
            pv_base_type = child.potentially_variable_base_class()
            if pv_base_type not in self:
                try:
                    child.__class__ = pv_base_type
                    self.register_dispatcher(visitor, child)
                finally:
                    child.__class__ = child_type
        elif (
            issubclass(child_type, _named_subexpression_types)
            or child_type is kernel.expression.noclone
        ):
            self[child_type] = self._before_named_expression
        else:
            self[child_type] = self._before_general_expression
        return self[child_type](visitor, child)

    @staticmethod
    def _before_general_expression(visitor, child):
        return True, None

    @staticmethod
    def _before_native_numeric(visitor, child):
        return False, (_CONSTANT, child)

    @staticmethod
    def _before_native_logical(visitor, child):
        return False, (
            _CONSTANT,
            InvalidNumber(
                child, f"{child!r} ({type(child).__name__}) is not a valid numeric type"
            ),
        )

    @staticmethod
    def _before_complex(visitor, child):
        return False, (_CONSTANT, complex_number_error(child, visitor, child))

    @staticmethod
    def _before_invalid(visitor, child):
        return False, (
            _CONSTANT,
            InvalidNumber(
                child, f"{child!r} ({type(child).__name__}) is not a valid numeric type"
            ),
        )

    @staticmethod
    def _before_string(visitor, child):
        return False, (
            _CONSTANT,
            InvalidNumber(
                child, f"{child!r} ({type(child).__name__}) is not a valid numeric type"
            ),
        )

    @staticmethod
    def _before_npv(visitor, child):
        try:
            return False, (
                _CONSTANT,
                check_constant(visitor.evaluate(child), child, visitor),
            )
        except (ValueError, ArithmeticError):
            return True, None

    @staticmethod
    def _before_param(visitor, child):
        return False, (_CONSTANT, check_constant(child.value, child, visitor))

    @staticmethod
    def _before_index_template(visitor, child):
        raise NotImplementedError(
            f"{visitor.__class__.__name__} can not handle expressions "
            f"containing {child.__class__} nodes"
        )

    @staticmethod
    def _before_indexed_expr(visitor, child):
        raise NotImplementedError(
            f"{visitor.__class__.__name__} can not handle expressions "
            f"containing {child.__class__} nodes"
        )

    @staticmethod
    def _before_indexed_param(visitor, child):
        raise NotImplementedError(
            f"{visitor.__class__.__name__} can not handle expressions "
            f"containing {child.__class__} nodes"
        )

    @staticmethod
    def _before_indexed_var(visitor, child):
        raise NotImplementedError(
            f"{visitor.__class__.__name__} can not handle expressions "
            f"containing {child.__class__} nodes"
        )

    @staticmethod
    def _before_component(visitor, child):
        raise NotImplementedError(
            f"{visitor.__class__.__name__} can not handle expressions "
            f"containing {child.__class__} nodes"
        )

    #
    # The following methods must be defined by derivative classes (along
    # with any other special-case handling they want to implement;
    # usually including handling for Monomial, Linear, and
    # ExternalFunction
    #

    # @staticmethod
    # def _before_var(visitor, child):
    #     pass

    # @staticmethod
    # def _before_named_expression(visitor, child):
    #     pass


class ExitNodeDispatcher(collections.defaultdict):
    """Dispatcher for handling the :class:`StreamBasedExpressionVisitor`
    `exitNode` callback

    This dispatcher implements a specialization of :class:`defaultdict`
    that supports automatic type registration.  As the identified
    callback is added to the dict, no type will incur the overhead of
    `register_dispatcher` more than once.

    Note that in this case, the client is expected to register all
    non-NPV expression types.  The auto-registration is designed to only
    handle two cases:
    - Auto-detection of user-defined Named Expression types
    - Automatic mappimg of NPV expressions to their equivalent non-NPV handlers
    - Automatic registration of derived expression types

    """

    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(None, *args, **kwargs)

    def __missing__(self, key):
        if type(key) is tuple:
            # Only lookup/cache argument-specific handlers for unary or
            # binary operators
            if len(key) <= 3:
                node_class = key[0]
                node_args = key[1:]
            else:
                node_class = key = key[0]
                if node_class in self:
                    return self[node_class]
        else:
            node_class = key
        bases = node_class.__mro__
        # Note: if we add an `etype`, then this special-case can be removed
        if (
            issubclass(node_class, _named_subexpression_types)
            or node_class is kernel.expression.noclone
        ):
            bases = [Expression]
        fcn = None
        for base_type in bases:
            if key is not node_class:
                if (base_type,) + node_args in self:
                    fcn = self[(base_type,) + node_args]
                    break
            if base_type in self:
                fcn = self[base_type]
                break
        if fcn is None:
            partial_matches = set(
                k[0] for k in self if type(k) is tuple and issubclass(node_class, k[0])
            )
            for base_type in node_class.__mro__:
                if node_class is not key:
                    key = (base_type,) + node_args
                if base_type in partial_matches:
                    raise DeveloperError(
                        f"Base expression key '{key}' not found when inserting "
                        f"dispatcher for node '{node_class.__name__}' while walking "
                        "expression tree."
                    )
            return self.unexpected_expression_type
        self[key] = fcn
        return fcn

    def unexpected_expression_type(self, visitor, node, *args):
        raise DeveloperError(
            f"Unexpected expression node type '{type(node).__name__}' "
            f"found while walking expression tree in {type(visitor).__name__}."
        )


def initialize_exit_node_dispatcher(exit_handlers):
    exit_dispatcher = {}
    for cls, handlers in exit_handlers.items():
        for args, fcn in handlers.items():
            if args is None:
                exit_dispatcher[cls] = fcn
            else:
                exit_dispatcher[(cls, *args)] = fcn
    return exit_dispatcher


def apply_node_operation(node, args):
    try:
        ans = node._apply_operation(args)
        if ans != ans and ans.__class__ is not InvalidNumber:
            ans = InvalidNumber(ans, "Evaluating '{node}' returned NaN")
        return ans
    except:
        exc_msg = str(sys.exc_info()[1])
        logger.warning(
            "Exception encountered evaluating expression "
            "'%s(%s)'\n\tmessage: %s\n\texpression: %s"
            % (node.name, ", ".join(map(str, args)), exc_msg, node)
        )
        if HALT_ON_EVALUATION_ERROR:
            raise
        return InvalidNumber(nan, exc_msg)


def complex_number_error(value, visitor, expr, node=""):
    msg = f'Pyomo {visitor.__class__.__name__} does not support complex numbers'
    cause = ' '.join(filter(None, ("Complex number returned from expression", node)))
    logger.warning(f"{cause}\n\tmessage: {msg}\n\texpression: {expr}")
    if HALT_ON_EVALUATION_ERROR:
        raise InvalidValueError(
            f'Pyomo {visitor.__class__.__name__} does not support complex numbers'
        )
    return InvalidNumber(value, cause)


def categorize_valid_components(
    model, active=True, sort=None, valid=set(), targets=set()
):
    """Walk model and check for valid component types

    This routine will walk the model and check all component types.
    Components types in the `valid` set are ignored, blocks with
    components in the `targets` set are collected, and all other
    component types are added to a dictionary of `unrecognized`
    components.

    A Component type may not appear in both `valid` and `targets` sets.

    Parameters
    ----------
    model: BlockData
        The model tree to walk

    active: True or None
        If True, only unrecognized active components are returned in the
        `uncategorized` dictionary.  Also, if True, only active Blocks
        are descended into.

    sort: bool or SortComponents
        The sorting flag to pass to the block walkers

    valid: Set[type]
        The set of "valid" component types.  These are ignored by the
        categorizer.

    targets: Set[type]
        The set of component types to "collect".  Blocks with components
        in the `targets` set will be returned in the `component_map`

    Returns
    -------
    component_map: Dict[type, List[BlockData]]
        A dict mapping component type to a list of block data
        objects that contain declared component of that type.

    unrecognized: Dict[type, List[ComponentData]]
        A dict mapping unrecognized component types to a (non-empty)
        list of component data objects found on the model.

    """
    assert active in (True, None)
    # Note: we assume every target component is valid but that we expect
    # there to be far mode valid components than target components.
    # Generate an error if a target is in the valid set (because the
    # valid set will preclude recording the block in the component_map)
    if any(ctype in valid for ctype in targets):
        ctypes = list(filter(valid.__contains__, targets))
        raise DeveloperError(
            f"categorize_valid_components: Cannot have component type {ctypes} in "
            "both the `valid` and `targets` sets"
        )
    unrecognized = {}
    component_map = {k: [] for k in targets}
    for block in model.block_data_objects(active=active, descend_into=True, sort=sort):
        local_ctypes = block.collect_ctypes(active=None, descend_into=False)
        for ctype in local_ctypes:
            if ctype in kernel.base._kernel_ctype_backmap:
                ctype = kernel.base._kernel_ctype_backmap[ctype]
            if ctype in valid:
                continue
            if ctype in targets:
                component_map[ctype].append(block)
                continue
            # TODO: we should rethink the definition of "active" for
            # Components that are not subclasses of ActiveComponent
            if (
                active
                and not issubclass(ctype, ActiveComponent)
                and not issubclass(ctype, kernel.base.ICategorizedObject)
            ):
                continue
            if ctype not in unrecognized:
                unrecognized[ctype] = []
            unrecognized[ctype].extend(
                block.component_data_objects(
                    ctype=ctype,
                    active=active,
                    descend_into=False,
                    sort=SortComponents.deterministic,
                )
            )
    return component_map, {k: v for k, v in unrecognized.items() if v}


def FileDeterminism_to_SortComponents(file_determinism):
    if file_determinism >= FileDeterminism.SORT_SYMBOLS:
        return SortComponents.ALPHABETICAL | SortComponents.SORTED_INDICES
    if file_determinism >= FileDeterminism.SORT_INDICES:
        return SortComponents.SORTED_INDICES
    if file_determinism >= FileDeterminism.ORDERED:
        return SortComponents.ORDERED_INDICES
    return SortComponents.UNSORTED


def initialize_var_map_from_column_order(model, config, var_map):
    column_order = config.column_order
    sorter = FileDeterminism_to_SortComponents(config.file_determinism)

    if column_order is None or column_order.__class__ is bool:
        if not column_order:
            column_order = None
    elif isinstance(column_order, ComponentMap):
        # The column order has historically has supported a ComponentMap of
        # component to position in addition to the simple list of
        # components.  Convert it to the simple list
        column_order = sorted(column_order, key=column_order.__getitem__)

    if column_order == True:
        column_order = model.component_data_objects(Var, descend_into=True, sort=sorter)
    elif config.file_determinism > FileDeterminism.ORDERED:
        # We will pre-gather the variables so that their order
        # matches the file_determinism flag.  This is a little
        # cumbersome, but is implemented this way for consistency
        # with the original NL writer.
        var_objs = model.component_data_objects(Var, descend_into=True, sort=sorter)
        if column_order is None:
            column_order = var_objs
        else:
            column_order = itertools.chain(column_order, var_objs)

    if column_order is not None:
        # Note that Vars that appear twice (e.g., through a
        # Reference) will be sorted with the FIRST occurrence.
        fill_in = ComponentSet()
        for var in column_order:
            if var.is_indexed():
                for _v in var.values(sorter):
                    if not _v.fixed:
                        var_map[id(_v)] = _v
            elif not var.fixed:
                pc = var.parent_component()
                if pc is not var and pc not in fill_in:
                    # For any VarData in an IndexedVar, remember the
                    # IndexedVar so that after all the VarData that the
                    # user has specified in the column ordering have
                    # been processed (and added to the var_map) we can
                    # go back and fill in any missing VarData from that
                    # IndexedVar.  This is needed because later when
                    # walking expressions we assume that any VarData
                    # that is not in the var_map will be the first
                    # VarData from its Var container (indexed or scalar).
                    fill_in.add(pc)
                var_map[id(var)] = var
        # Note that ComponentSet iteration is deterministic, and
        # re-inserting _v into the var_map will not change the ordering
        # for any pre-existing variables
        for pc in fill_in:
            for _v in pc.values(sorter):
                if not _v.fixed:
                    var_map[id(_v)] = _v
    return var_map


def row_order2row_map(config):
    """Convert a row_order (list or ComponentMap) into a dict mapping
    constraint id -> row index. Returns an empty dict if no ordering."""
    row_order = config.row_order
    if row_order is None or isinstance(row_order, bool):
        return {}

    sorter = FileDeterminism_to_SortComponents(config.file_determinism)

    if isinstance(row_order, ComponentMap):
        # Convert ComponentMap to sorted list based on its values
        row_order = sorted(row_order, key=row_order.__getitem__)

    row_map = {}
    for con in row_order:
        if con.is_indexed():
            for c in con.values(sorter):
                row_map[id(c)] = c
        else:
            row_map[id(con)] = con

    # Map the implicit dict ordering to an explicit 0..n ordering
    return {_id: i for i, _id in enumerate(row_map)}


def ordered_active_constraints(model, config):
    sorter = FileDeterminism_to_SortComponents(config.file_determinism)
    constraints = model.component_data_objects(Constraint, active=True, sort=sorter)

    row_map = row_order2row_map(config)
    if not row_map:
        return constraints
    # sorted() is stable (per Python docs), so we can let all
    # unspecified rows have a row number one bigger than the
    # number of rows specified by the user ordering.
    _n = len(row_map)
    _row_getter = row_map.get
    return sorted(constraints, key=lambda x: _row_getter(id(x), _n))


class VarRecorder:
    def __init__(self, var_map, sorter):
        self.var_map = var_map
        self.sorter = sorter

    def add(self, var):
        # We always add all indices to the var_map at once so that
        # we can honor deterministic ordering of unordered sets
        # (because the user could have iterated over an unordered
        # set when constructing an expression, thereby altering the
        # order in which we would see the variables)
        vm = self.var_map
        try:
            _iter = var.parent_component().values(self.sorter)
        except AttributeError:
            # Note that this only works for the AML, as kernel does not
            # provide a parent_component()
            _iter = (var,)
        for v in _iter:
            if not v.fixed:
                vm[id(v)] = v


class OrderedVarRecorder:
    def __init__(self, var_map, var_order, sorter):
        self.var_map = var_map
        self.var_order = var_order
        self.sorter = sorter

    def add(self, var):
        # We always add all indices to the var_map at once so that
        # we can honor deterministic ordering of unordered sets
        # (because the user could have iterated over an unordered
        # set when constructing an expression, thereby altering the
        # order in which we would see the variables)
        vm = self.var_map
        vo = self.var_order
        try:
            _iter = var.parent_component().values(self.sorter)
        except AttributeError:
            # Note that this only works for the AML, as kernel does not
            # provide a parent_component()
            _iter = (var,)
        for i, v in enumerate(_iter, start=len(vo)):
            vid = id(v)
            vo[vid] = i
            if not v.fixed:
                vm[vid] = v


class TemplateVarRecorder:
    def __init__(self, var_map, sorter):
        self.var_map = var_map
        self._var_order = None
        self.sorter = sorter
        self.env = {None: 0}
        self.symbolmap = EXPR.SymbolMap(NumericLabeler('x'))
        if var_map:
            # If the user provided an initial var_map, we want to honor
            # that ordering.  This means we need to both initialize the
            # env dict with all the Vars referenced, PLUS fill in any
            # additional vars that we would have indexed/recorded in
            # add()
            next_i = len(var_map)
            for i, v in enumerate(list(var_map.values())):
                var_comp = v.parent_component()
                name = self.symbolmap.getSymbol(var_comp)
                ve = self.env.get(name, None)
                if ve is None:
                    ve = self.env[name] = {}
                    for idx, vdata in var_comp.items():
                        vid = id(vdata)
                        if vid not in var_map:
                            var_map[vid] = v
                            ve[idx] = next_i
                            next_i += 1
                ve[v.index()] = i

    @property
    def var_order(self):
        if self._var_order is None:
            self._var_order = {vid: i for i, vid in enumerate(self.var_map)}
        return self._var_order

    def add(self, var):
        # Note: the following is mostly a copy of
        # LinearBeforeChildDispatcher.record_var, but with extra
        # handling to update the env in the same loop
        var_comp = var.parent_component()
        # Double-check that the component has not already been processed
        # (through an individual var data)
        name = self.symbolmap.getSymbol(var_comp)
        if name in self.env:
            return

        # We always add all indices to the var_map at once so that
        # we can honor deterministic ordering of unordered sets
        # (because the user could have iterated over an unordered
        # set when constructing an expression, thereby altering the
        # order in which we would see the variables)
        vm = self.var_map
        ve = self.env[name] = {}
        try:
            _iter = var_comp.items(self.sorter)
        except AttributeError:
            # Note that this only works for the AML, as kernel does not
            # provide a parent_component()
            _iter = (None, var)
        if self._var_order is None:
            for i, (idx, v) in enumerate(_iter, start=len(vm)):
                vm[id(v)] = v
                ve[idx] = i
        else:
            vo = self._var_order
            for i, (idx, v) in enumerate(_iter, start=len(vm)):
                vid = id(v)
                vm[vid] = v
                ve[idx] = i
                vo[vid] = i


# Copied from cpxlp.py:
# Keven Hunter made a nice point about using %.16g in his attachment
# to ticket #4319. I am adjusting this to %.17g as this mocks the
# behavior of using %r (i.e., float('%r'%<number>) == <number>) with
# the added benefit of outputting (+/-). The only case where this
# fails to mock the behavior of %r is for large (long) integers (L),
# which is a rare case to run into and is probably indicative of
# other issues with the model.
# *** NOTE ***: If you use 'r' or 's' here, it will break code that
#               relies on using '%+' before the formatting character
#               and you will need to go add extra logic to output
#               the number's sign.
_ftoa_precision_str = '%.17g'


def ftoa(val, parenthesize_negative_values=False):
    if val is None:
        return val
    #
    # Basic checking, including conversion of *fixed* Pyomo types to floats
    if type(val) in native_numeric_types:
        _val = val
    else:
        if is_fixed(val):
            _val = value(val)
        else:
            raise ValueError(
                "Converting non-fixed bound or value to string: %s" % (val,)
            )
    #
    # Convert to string
    a = _ftoa_precision_str % _val
    #
    # Remove unnecessary least significant digits.  While not strictly
    # necessary, this helps keep the emitted string consistent between
    # python versions by simplifying things like "1.0000000000001" to
    # "1".
    i = len(a) - 1
    if i:
        try:
            while float(a[:i]) == _val:
                i -= 1
        except:
            pass
    i += 1
    #
    # It is important to issue a warning if the conversion loses
    # precision (as the emitted model is not exactly what the user
    # specified)
    if i == len(a) and float(a) != _val:
        logger.warning("Converting %s to string resulted in loss of precision" % val)
    #
    if parenthesize_negative_values and a[0] == '-':
        return '(' + a[:i] + ')'
    else:
        return a[:i]


def check_constant(ans, obj, visitor):
    # [ESJ 10/15/25]: The only reason this takes the visitor as an
    # argument is to pass it to the complex_number_error function, and
    # the only reason it needs it is to get the name of the class. But
    # it is public, so I'm not changing anything about that at the
    # moment.
    if ans.__class__ not in native_numeric_types:
        # None can be returned from uninitialized Var/Param objects
        if ans is None:
            return InvalidNumber(
                None, f"'{obj}' evaluated to a nonnumeric value '{ans}'"
            )
        if ans.__class__ is InvalidNumber:
            return ans
        elif ans.__class__ in native_complex_types:
            return complex_number_error(ans, visitor, obj)
        else:
            # It is possible to get other non-numeric types.  Most
            # common are bool and 1-element numpy.array().  We will
            # attempt to convert the value to a float before
            # proceeding.
            #
            # Note that as of NumPy 1.25, blindly casting a
            # 1-element ndarray to a float will generate a
            # deprecation warning.  We will explicitly test for
            # that, but want to do the test without triggering the
            # numpy import
            for cls in ans.__class__.__mro__:
                if cls.__name__ == 'ndarray' and cls.__module__ == 'numpy':
                    if len(ans) == 1:
                        ans = ans[0]
                    break
            # TODO: we should check bool and warn/error (while bool is
            # convertible to float in Python, they have very
            # different semantic meanings in Pyomo).
            try:
                ans = float(ans)
            except:
                return InvalidNumber(
                    ans, f"'{obj}' evaluated to a nonnumeric value '{ans}'"
                )
    if ans != ans:
        return InvalidNumber(nan, f"'{obj}' evaluated to a nonnumeric value '{ans}'")
    return ans



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from pyomo.opt.base import (
    check_available_solvers,
    convert,
    convert_problem,
    error,
    formats,
    guess_format,
    opt_config,
    solvers,
    AbstractProblemWriter,
    AbstractResultsReader,
    BranchDirection,
    ConverterError,
    OptSolver,
    ProblemFormat,
    ReaderFactory,
    ResultsFormat,
    SolverFactory,
    UnknownSolver,
    WriterFactory,
)

from pyomo.opt.results import (
    container,
    problem,
    solution,
    ListContainer,
    MapContainer,
    UndefinedData,
    undefined,
    ignore,
    SolverStatus,
    TerminationCondition,
    check_optimal_termination,
    assert_optimal_termination,
    ProblemSense,
    SolutionStatus,
    Solution,
    results_,
    SolverResults,
)

from pyomo.opt.problem import ampl, AmplModel

from pyomo.opt.parallel import (
    manager,
    async_solver,
    local,
    SolverManagerFactory,
    AsynchronousSolverManager,
)

from pyomo.common.deprecation import relocated_module_attribute

for _attr in ('ScalarData', 'ScalarType', 'default_print_options'):
    relocated_module_attribute(
        _attr, 'pyomo.opt.results.container.' + _attr, version='6.0'
    )
del _attr
del relocated_module_attribute



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.autoslots import AutoSlots
from pyomo.core.base.block import BlockData, IndexedBlock
from pyomo.core.base.global_set import UnindexedComponent_index, UnindexedComponent_set


class _TransformedDisjunctData(BlockData):
    __slots__ = ('_src_disjunct',)
    __autoslot_mappers__ = {'_src_disjunct': AutoSlots.weakref_mapper}

    @property
    def src_disjunct(self):
        return None if self._src_disjunct is None else self._src_disjunct()

    def __init__(self, component):
        BlockData.__init__(self, component)
        # pointer to the Disjunct whose transformation block this is.
        self._src_disjunct = None


class _TransformedDisjunct(IndexedBlock):
    _ComponentDataClass = _TransformedDisjunctData



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
from pyomo.environ import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    Suffix,
    NonNegativeIntegers,
    NonNegativeReals,
    value,
)
from pyomo.opt import SolverFactory, TerminationCondition

knitroampl_available = SolverFactory('knitroampl').available(False)


@unittest.skipIf(not knitroampl_available, "The 'knitroampl' command is not available")
@unittest.pytest.mark.solver("knitroampl")
class TestKNITROAMPLInterface(unittest.TestCase):
    def test_infeasible_lp(self):
        with SolverFactory('knitroampl') as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeReals)
            model.C1 = Constraint(expr=model.X == 1)
            model.C2 = Constraint(expr=model.X == 2)
            model.Obj = Objective(expr=model.X)

            results = opt.solve(model)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.infeasible
            )

    def test_unbounded_lp(self):
        with SolverFactory('knitroampl') as opt:
            model = ConcreteModel()
            model.X = Var()
            model.Obj = Objective(expr=model.X)

            results = opt.solve(model)

            self.assertIn(
                results.solver.termination_condition,
                (
                    TerminationCondition.unbounded,
                    TerminationCondition.infeasibleOrUnbounded,
                ),
            )

    def test_optimal_lp(self):
        with SolverFactory('knitroampl') as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeReals)
            model.C1 = Constraint(expr=model.X >= 2.5)
            model.Obj = Objective(expr=model.X)

            results = opt.solve(model, load_solutions=True)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )
            self.assertAlmostEqual(value(model.X), 2.5)

    def test_get_duals_lp(self):
        with SolverFactory('knitroampl') as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeReals)
            model.Y = Var(within=NonNegativeReals)

            model.C1 = Constraint(expr=2 * model.X + model.Y >= 8)
            model.C2 = Constraint(expr=model.X + 3 * model.Y >= 6)

            model.Obj = Objective(expr=model.X + model.Y)

            results = opt.solve(model, suffixes=['dual'], load_solutions=False)

            model.dual = Suffix(direction=Suffix.IMPORT)
            model.solutions.load_from(results)

            self.assertAlmostEqual(model.dual[model.C1], 0.4)
            self.assertAlmostEqual(model.dual[model.C2], 0.2)

    def test_infeasible_mip(self):
        with SolverFactory('knitroampl') as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeIntegers)
            model.C1 = Constraint(expr=model.X == 1)
            model.C2 = Constraint(expr=model.X == 2)
            model.Obj = Objective(expr=model.X)

            results = opt.solve(model)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.infeasible
            )

    def test_optimal_mip(self):
        with SolverFactory('knitroampl') as opt:
            model = ConcreteModel()
            model.X = Var(within=NonNegativeIntegers)
            model.C1 = Constraint(expr=model.X >= 2.5)
            model.Obj = Objective(expr=model.X)

            results = opt.solve(model, load_solutions=True)

            self.assertEqual(
                results.solver.termination_condition, TerminationCondition.optimal
            )
            self.assertAlmostEqual(value(model.X), 3)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Pyomo: Python Optimization Modeling Objects

Pyomo provides Python packages for formulating and solving complex
optimization applications.  Most of Pyomo's packages rely on externally
built optimization solvers.

pyomo.version provides a mechanism for managing stuff that is related to
releases of the entire Pyomo software.
"""

from pyomo.version.info import version, version_info, __version__



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import collections.abc

from pyomo.core.kernel.tuple_container import TupleContainer

logger = logging.getLogger('pyomo.core')


class ListContainer(TupleContainer, collections.abc.MutableSequence):
    """
    A partial implementation of the IHomogeneousContainer
    interface that provides list-like storage functionality.

    Complete implementations need to set the _ctype property
    at the class level and initialize the remaining
    ICategorizedObject attributes during object creation. If
    using __slots__, a slot named "_data" must be included.

    Note that this implementation allows nested storage of
    other ICategorizedObjectContainer implementations that
    are defined with the same ctype.
    """

    __slots__ = ()

    def __init__(self, *args):
        self._data = []
        self._init(args)

    #
    # Define the MutableSequence abstract methods
    #

    def __setitem__(self, i, item):
        if item.ctype is self.ctype:
            if item._parent is None:
                # be sure the current object is properly
                # removed
                logger.warning(
                    "Implicitly replacing the entry %s (type=%s) "
                    "with a new object (type=%s). This is usually "
                    "indicative of a modeling error. To avoid this "
                    "warning, delete the original object from the "
                    "container before assigning a new object."
                    % (
                        self[i].name,
                        self[i].__class__.__name__,
                        item.__class__.__name__,
                    )
                )
                self._data[i]._clear_parent_and_storage_key()
                item._update_parent_and_storage_key(self, i)
                self._data[i] = item
                return
            elif self._data[i] is item:
                # a very special case that makes sense to handle
                # because the implied order should be: (1) delete
                # the object at the current index, (2) insert the
                # the new object. This performs both without any
                # actions, but it is an extremely rare case, so
                # it should go last.
                return
            # see note about allowing objects to live in more than
            # one container
            raise ValueError(
                "Invalid assignment to %s type with name '%s' "
                "at index %s. A parent container has already been "
                "assigned to the object being inserted: %s"
                % (self.__class__.__name__, self.name, i, item.parent.name)
            )
        else:
            raise TypeError(
                "Invalid assignment to type %s with index %s. "
                "The object being inserted has the wrong "
                "category type: %s" % (self.__class__.__name__, i, item.ctype)
            )

    def insert(self, i, item):
        """S.insert(index, object) -- insert object before index"""
        self._insert(i, item)

    def __delitem__(self, i):
        self._data[i]._clear_parent_and_storage_key()
        del self._data[i]

    #
    # Override a few default implementations on Sequence
    #
    # We want to avoid issues with parent ownership
    # as we shuffle items

    def reverse(self):
        """S.reverse() -- reverse *IN PLACE*"""
        n = len(self)
        data = self._data
        for i in range(n // 2):
            data[i], data[n - i - 1] = data[n - i - 1], data[i]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Re-implementation of eight-process problem.

Re-implementation of Duran example 3 superstructure synthesis problem in Pyomo
Author: Qi Chen <https://github.com/qtothec>.

Ref:
    SELECT OPTIMAL PROCESS FROM WITHIN GIVEN SUPERSTRUCTURE.
    MARCO DURAN , PH.D. THESIS (EX3) , 1984.
    CARNEGIE-MELLON UNIVERSITY , PITTSBURGH , PA.

The expected optimal solution value is 68.0.

(original problem, my implementation may vary)
    Problem type:    convex MINLP
            size:    8  binary variables
                    26  continuous variables
                    32  constraints

Pictoral representation can be found on
page 969 of Turkay & Grossmann, 1996.
http://dx.doi.org/10.1016/0098-1354(95)00219-7

"""

from pyomo.environ import (
    Binary,
    ConcreteModel,
    Constraint,
    NonNegativeReals,
    Objective,
    Param,
    RangeSet,
    Var,
    exp,
    minimize,
)
from pyomo.common.collections import ComponentMap


class EightProcessFlowsheet(ConcreteModel):
    """Flowsheet for the 8 process problem."""

    def __init__(self, convex=True, *args, **kwargs):
        """Create the flowsheet."""
        kwargs.setdefault('name', 'DuranEx3')
        super(EightProcessFlowsheet, self).__init__(*args, **kwargs)
        m = self

        """Set declarations"""
        I = m.I = RangeSet(2, 25, doc='process streams')
        J = m.J = RangeSet(1, 8, doc='process units')
        m.PI = RangeSet(1, 4, doc='integer constraints')
        m.DS = RangeSet(1, 4, doc='design specifications')
        """
        1: Unit 8
        2: Unit 8
        3: Unit 4
        4: Unit 4
        """
        m.MB = RangeSet(1, 7, doc='mass balances')
        """Material balances:
        1: 4-6-7
        2: 3-5-8
        3: 4-5
        4: 1-2
        5: 1-2-3
        6: 6-7-4
        7: 6-7
        """

        """Parameter and initial point declarations"""
        # FIXED COST INVESTMENT COEFF FOR PROCESS UNITS
        # Format: process #: cost
        fixed_cost = {1: 5, 2: 8, 3: 6, 4: 10, 5: 6, 6: 7, 7: 4, 8: 5}
        CF = m.CF = Param(J, initialize=fixed_cost)

        # VARIABLE COST COEFF FOR PROCESS UNITS - STREAMS
        # Format: stream #: cost
        variable_cost = {
            3: -10,
            5: -15,
            9: -40,
            19: 25,
            21: 35,
            25: -35,
            17: 80,
            14: 15,
            10: 15,
            2: 1,
            4: 1,
            18: -65,
            20: -60,
            22: -80,
        }
        CV = m.CV = Param(I, initialize=variable_cost, default=0)

        # initial point information for equipment selection (for each NLP
        # subproblem)
        initY = {
            'sub1': {1: 1, 2: 0, 3: 1, 4: 1, 5: 0, 6: 0, 7: 1, 8: 1},
            'sub2': {1: 0, 2: 1, 3: 1, 4: 1, 5: 0, 6: 1, 7: 0, 8: 1},
            'sub3': {1: 1, 2: 0, 3: 1, 4: 0, 5: 1, 6: 0, 7: 0, 8: 1},
        }
        # initial point information for stream flows
        initX = {
            2: 2,
            3: 1.5,
            4: 0,
            5: 0,
            6: 0.75,
            7: 0.5,
            8: 0.5,
            9: 0.75,
            10: 0,
            11: 1.5,
            12: 1.34,
            13: 2,
            14: 2.5,
            15: 0,
            16: 0,
            17: 2,
            18: 0.75,
            19: 2,
            20: 1.5,
            21: 0,
            22: 0,
            23: 1.7,
            24: 1.5,
            25: 0.5,
        }

        """Variable declarations"""
        # BINARY VARIABLE DENOTING EXISTENCE-NONEXISTENCE
        Y = m.Y = Var(J, domain=Binary, initialize=initY['sub1'])
        # FLOWRATES OF PROCESS STREAMS
        X = m.X = Var(I, domain=NonNegativeReals, initialize=initX)
        # OBJECTIVE FUNCTION CONSTANT TERM
        CONSTANT = m.constant = Param(initialize=122.0)

        """Constraint definitions"""
        # INPUT-OUTPUT RELATIONS FOR process units 1 through 8
        m.inout3 = Constraint(expr=1.5 * m.X[9] + m.X[10] == m.X[8])
        m.inout4 = Constraint(expr=1.25 * (m.X[12] + m.X[14]) == m.X[13])
        m.inout5 = Constraint(expr=m.X[15] == 2 * m.X[16])
        if convex:
            m.inout1 = Constraint(expr=exp(m.X[3]) - 1 <= m.X[2])
            m.inout2 = Constraint(expr=exp(m.X[5] / 1.2) - 1 <= m.X[4])
            m.inout7 = Constraint(expr=exp(m.X[22]) - 1 <= m.X[21])
            m.inout8 = Constraint(expr=exp(m.X[18]) - 1 <= m.X[10] + m.X[17])
            m.inout6 = Constraint(expr=exp(m.X[20] / 1.5) - 1 <= m.X[19])
        else:
            m.inout1 = Constraint(expr=exp(m.X[3]) - 1 == m.X[2])
            m.inout2 = Constraint(expr=exp(m.X[5] / 1.2) - 1 == m.X[4])
            m.inout7 = Constraint(expr=exp(m.X[22]) - 1 == m.X[21])
            m.inout8 = Constraint(expr=exp(m.X[18]) - 1 == m.X[10] + m.X[17])
            m.inout6 = Constraint(expr=exp(m.X[20] / 1.5) - 1 == m.X[19])

        # Mass balance equations
        m.massbal1 = Constraint(expr=m.X[13] == m.X[19] + m.X[21])
        m.massbal2 = Constraint(expr=m.X[17] == m.X[9] + m.X[16] + m.X[25])
        m.massbal3 = Constraint(expr=m.X[11] == m.X[12] + m.X[15])
        m.massbal4 = Constraint(expr=m.X[3] + m.X[5] == m.X[6] + m.X[11])
        m.massbal5 = Constraint(expr=m.X[6] == m.X[7] + m.X[8])
        m.massbal6 = Constraint(expr=m.X[23] == m.X[20] + m.X[22])
        m.massbal7 = Constraint(expr=m.X[23] == m.X[14] + m.X[24])

        # process specifications
        m.specs1 = Constraint(expr=m.X[10] <= 0.8 * m.X[17])
        m.specs2 = Constraint(expr=m.X[10] >= 0.4 * m.X[17])
        m.specs3 = Constraint(expr=m.X[12] <= 5 * m.X[14])
        m.specs4 = Constraint(expr=m.X[12] >= 2 * m.X[14])

        # Logical constraints (big-M) for each process.
        # These allow for flow iff unit j exists
        m.logical1 = Constraint(expr=m.X[2] <= 10 * m.Y[1])
        m.logical2 = Constraint(expr=m.X[4] <= 10 * m.Y[2])
        m.logical3 = Constraint(expr=m.X[9] <= 10 * m.Y[3])
        m.logical4 = Constraint(expr=m.X[12] + m.X[14] <= 10 * m.Y[4])
        m.logical5 = Constraint(expr=m.X[15] <= 10 * m.Y[5])
        m.logical6 = Constraint(expr=m.X[19] <= 10 * m.Y[6])
        m.logical7 = Constraint(expr=m.X[21] <= 10 * m.Y[7])
        m.logical8 = Constraint(expr=m.X[10] + m.X[17] <= 10 * m.Y[8])

        # pure integer constraints
        m.pureint1 = Constraint(expr=m.Y[1] + m.Y[2] == 1)
        m.pureint2 = Constraint(expr=m.Y[4] + m.Y[5] <= 1)
        m.pureint3 = Constraint(expr=m.Y[6] + m.Y[7] - m.Y[4] == 0)
        m.pureint4 = Constraint(expr=m.Y[3] - m.Y[8] <= 0)

        """Cost (objective) function definition"""
        m.objective = Objective(
            expr=sum(Y[j] * CF[j] for j in J) + sum(X[i] * CV[i] for i in I) + CONSTANT,
            sense=minimize,
        )

        """Bound definitions"""
        # x (flow) upper bounds
        # x_ubs = {3: 2, 5: 2, 9: 2, 10: 1, 14: 1, 17: 2, 19: 2, 21: 2, 25: 3}
        x_ubs = {
            2: 10,
            3: 2,
            4: 10,
            5: 2,
            9: 2,
            10: 1,
            14: 1,
            17: 2,
            18: 10,
            19: 2,
            20: 10,
            21: 2,
            22: 10,
            25: 3,
        }  # add bounds for variables in nonlinear constraints
        for i, x_ub in x_ubs.items():
            X[i].setub(x_ub)
        m.optimal_value = 68.0097
        m.optimal_solution = ComponentMap()
        m.optimal_solution[m.X[2]] = 0.0
        m.optimal_solution[m.X[3]] = 0.0
        m.optimal_solution[m.X[4]] = 4.294490050470028
        m.optimal_solution[m.X[5]] = 1.9999999999999998
        m.optimal_solution[m.X[6]] = 0.6666666666666665
        m.optimal_solution[m.X[7]] = 0.1979983174202069
        m.optimal_solution[m.X[8]] = 0.4686683492464596
        m.optimal_solution[m.X[9]] = 0.0
        m.optimal_solution[m.X[10]] = 0.4686683492464596
        m.optimal_solution[m.X[11]] = 1.3333333333333333
        m.optimal_solution[m.X[12]] = 1.3333333333333333
        m.optimal_solution[m.X[13]] = 2.0
        m.optimal_solution[m.X[14]] = 0.26666666666666666
        m.optimal_solution[m.X[15]] = 0.0
        m.optimal_solution[m.X[16]] = 0.0
        m.optimal_solution[m.X[17]] = 0.5858354365580745
        m.optimal_solution[m.X[18]] = 0.720035498875516
        m.optimal_solution[m.X[19]] = 2.0
        m.optimal_solution[m.X[20]] = 1.6479184330021648
        m.optimal_solution[m.X[21]] = 0.0
        m.optimal_solution[m.X[22]] = 0.0
        m.optimal_solution[m.X[23]] = 1.6479184330021648
        m.optimal_solution[m.X[24]] = 1.3812517663354982
        m.optimal_solution[m.X[25]] = 0.5858354365580745
        m.optimal_solution[m.Y[1]] = 0.0
        m.optimal_solution[m.Y[2]] = 1.0
        m.optimal_solution[m.Y[3]] = 0.0
        m.optimal_solution[m.Y[4]] = 1.0
        m.optimal_solution[m.Y[5]] = 0.0
        m.optimal_solution[m.Y[6]] = 1.0
        m.optimal_solution[m.Y[7]] = 0.0
        m.optimal_solution[m.Y[8]] = 1.0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Elements of a Model
#
# TestSimpleCon                Class for testing single constraint
# TestArrayCon                Class for testing array of constraint
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

import pyomo.common.unittest as unittest

from pyomo.environ import (
    AbstractModel,
    Param,
    Block,
    Set,
    Var,
    RangeSet,
    Constraint,
    Connector,
    value,
)


class Test(unittest.TestCase):
    def test_nonindexed_block_immutable_param(self):
        model = AbstractModel()

        def _b_rule(b):
            b.A = Param(initialize=2.0)

        model.B = Block(rule=_b_rule)

        instance = model.create_instance()

        self.assertEqual(value(instance.B.A), 2.0)

    def test_nonindexed_block_mutable_param(self):
        model = AbstractModel()

        def _b_rule(b):
            b.A = Param(initialize=2.0, mutable=True)

        model.B = Block(rule=_b_rule)

        instance = model.create_instance()
        self.assertEqual(value(instance.B.A), 2.0)

        instance.B.A = 4.0
        self.assertEqual(value(instance.B.A), 4.0)

    def test_indexed_block_immutable_param(self):
        model = AbstractModel()
        model.A = RangeSet(2)

        def _b_rule(b, id):
            b.A = Param(initialize=id)

        model.B = Block(model.A, rule=_b_rule)

        instance = model.create_instance()

        self.assertEqual(value(instance.B[1].A), 1)
        self.assertEqual(value(instance.B[2].A), 2)

    def test_indexed_block_mutable_param(self):
        model = AbstractModel()
        model.A = RangeSet(2)

        def _b_rule(b, id):
            b.A = Param(initialize=id, mutable=True)

        model.B = Block(model.A, rule=_b_rule)

        instance = model.create_instance()

        self.assertEqual(value(instance.B[1].A), 1)
        self.assertEqual(value(instance.B[2].A), 2)

        instance.B[1].A = 4.0
        self.assertEqual(value(instance.B[1].A), 4.0)

    def test_create_from_dict(self):
        model = AbstractModel()
        model.A = RangeSet(2)

        def _b_rule(b, id):
            b.S = Set()
            b.P = Param()
            b.Q = Param(b.S)

        model.B = Block(model.A, rule=_b_rule)

        instance = model.create_instance(
            {
                None: {
                    'B': {
                        1: {
                            'S': {None: ['a', 'b', 'c']},
                            'P': {None: 4},
                            'Q': {('a',): 1, ('b',): 2, ('c',): 3},
                        },
                        2: {'S': {None: []}, 'P': {None: 3}},
                    }
                }
            }
        )

        self.assertEqual(set(instance.B[1].S), set(['a', 'b', 'c']))
        self.assertEqual(value(instance.B[1].P), 4)
        self.assertEqual(value(instance.B[1].Q['a']), 1)
        self.assertEqual(value(instance.B[1].Q['b']), 2)
        self.assertEqual(value(instance.B[1].Q['c']), 3)
        self.assertEqual(value(instance.B[2].P), 3)

    def test_expand_connector(self):
        model = AbstractModel()
        model.A = Set()

        def _b_rule(b, id):
            b.X = Var()
            b.PORT = Connector()
            b.PORT.add(b.X)

        model.B = Block(model.A, rule=_b_rule)

        def _c_rule(m, a):
            return m.B[a].PORT == m.B[(a + 1) % 2].PORT

        model.C = Constraint(model.A, rule=_c_rule)

        instance = model.create_instance({None: {'A': {None: [0, 1]}}})

        # FIXME: Not sure what to assert here, but at the moment this throws an error anyways.

    def test_len(self):
        model = AbstractModel()
        model.b = Block()

        # TODO: I think the correct answer should be zero before
        # construction
        self.assertEqual(len(model.b), 1)
        inst = model.create_instance()
        self.assertEqual(len(inst.b), 1)

    def test_none_key(self):
        model = AbstractModel()
        model.b = Block()

        inst = model.create_instance()
        self.assertEqual(id(inst.b), id(inst.b[None]))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

from collections import namedtuple
from pyomo.core.expr.numvalue import value as pyo_value
from pyomo.contrib.mpc.data.get_cuid import get_indexed_cuid
from pyomo.contrib.mpc.data.dynamic_data_base import _is_iterable, _DynamicDataBase
from pyomo.contrib.mpc.data.scalar_data import ScalarData
from pyomo.contrib.mpc.data.find_nearest_index import (
    find_nearest_index,
    find_nearest_interval_index,
)

IntervalDataTuple = namedtuple("IntervalTuple", ["data", "intervals"])


def assert_disjoint_intervals(intervals):
    """
    This function takes intervals in the form of tuples and makes sure
    that they are disjoint.

    Arguments
    ---------
    intervals: iterable
        Iterable of tuples, each containing the low and high values of an
        interval.

    """
    intervals = list(sorted(intervals))
    for i, (lo, hi) in enumerate(intervals):
        if not lo <= hi:
            raise RuntimeError(
                "Lower endpoint of interval is higher than upper endpoint"
            )
        if i != 0:
            prev_lo, prev_hi = intervals[i - 1]
            if not prev_hi <= lo:
                raise RuntimeError(
                    "Intervals %s and %s are not disjoint"
                    % ((prev_lo, prev_hi), (lo, hi))
                )


class IntervalData(_DynamicDataBase):
    def __init__(self, data, intervals, time_set=None, context=None):
        intervals = list(intervals)
        if not intervals == list(sorted(intervals)):
            raise RuntimeError("Intervals are not sorted in increasing order.")
        assert_disjoint_intervals(intervals)
        self._intervals = intervals

        # First make sure provided lists of variable data have the
        # same lengths as the provided time list.
        for key, data_list in data.items():
            if len(data_list) != len(intervals):
                raise ValueError(
                    "Data lists must have same length as time. "
                    "Length of time is %s while length of data for "
                    "key %s is %s." % (len(intervals), key, len(data_list))
                )
        super().__init__(data, time_set=time_set, context=context)

    def __eq__(self, other):
        if isinstance(other, IntervalData):
            return (
                self._data == other.get_data()
                and self._intervals == other.get_intervals()
            )
        else:
            raise TypeError(
                "%s and %s are not comparable" % (self.__class__, other.__class__)
            )

    def get_intervals(self):
        return self._intervals

    def get_data_at_interval_indices(self, indices):
        # NOTE: Much of this code is repeated from TimeSeriesData.
        # TODO: Find some way to consolidate.
        if _is_iterable(indices):
            index_list = list(sorted(indices))
            interval_list = [self._intervals[i] for i in indices]
            data = {
                cuid: [values[idx] for idx in index_list]
                for cuid, values in self._data.items()
            }
            time_set = self._orig_time_set
            return IntervalData(data, interval_list, time_set=time_set)
        else:
            return ScalarData(
                {cuid: values[indices] for cuid, values in self._data.items()}
            )

    # TODO: get_data_at_interval, get_data_at_time
    def get_data_at_time(self, time, tolerance=None, prefer_left=True):
        if not _is_iterable(time):
            index = find_nearest_interval_index(
                self._intervals, time, tolerance=tolerance, prefer_left=prefer_left
            )
            if index is None:
                raise RuntimeError(
                    "Time point %s not found in an interval within"
                    " tolerance %s" % (time, tolerance)
                )
        else:
            raise RuntimeError(
                "get_data_at_time is not supported with multiple time points"
                " for IntervalData. To sample the piecewise-constant data at"
                " particular time points, please use interval_to_series from"
                " pyomo.contrib.mpc.data.convert"
            )
        return self.get_data_at_interval_indices(index)

    def to_serializable(self):
        """
        Convert to json-serializable object.

        """
        intervals = self._intervals
        data = {
            str(cuid): [pyo_value(val) for val in values]
            for cuid, values in self._data.items()
        }
        return IntervalDataTuple(data, intervals)

    def concatenate(self, other, tolerance=0.0):
        """
        Extend interval list and variable data lists with the intervals
        and variable values in the provided IntervalData

        """
        other_intervals = other.get_intervals()
        intervals = self._intervals
        if len(other_intervals) == 0:
            return
        if other_intervals[0][0] < intervals[-1][1] + tolerance:
            # First point of new intervals is less than (within
            # tolerance)
            raise ValueError(
                "Initial time point of target, %s, is not greater than"
                " final time point of source, %s, within tolerance %s."
                % (other_time[0][0], time[-1][1], tolerance)
            )
        self._intervals.extend(other_intervals)

        data = self._data
        other_data = other.get_data()
        for cuid, values in data.items():
            # We assume that other contains all the cuids in self.
            # We make no assumption the other way around.
            values.extend(other_data[cuid])

    def shift_time_points(self, offset):
        """
        Apply an offset to stored time points.

        """
        # Note that this is different from what we are doing in
        # shift_values_by_time in the helper class.
        self._intervals = [(lo + offset, hi + offset) for lo, hi in self._intervals]

    def extract_variables(self, variables, context=None, copy_values=False):
        """
        Only keep variables specified.

        """
        if copy_values:
            raise NotImplementedError(
                "extract_variables with copy_values=True has not been"
                " implemented by %s" % self.__class__
            )
        data = {}
        if not isinstance(variables, (list, tuple)):
            # If variables is not a sequence and is instead a slice (or
            # indexed variable), we get either a confusing error message
            # or a lot of repeated work.
            raise TypeError("extract_values only accepts a list or tuple of variables")
        for var in variables:
            cuid = get_indexed_cuid(var, (self._orig_time_set,), context=context)
            data[cuid] = self._data[cuid]
        return IntervalData(data, self._intervals, time_set=self._orig_time_set)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import logging

from pyomo.opt import OptSolver, SolverFactory

logger = logging.getLogger('pyomo.solvers')


@SolverFactory.register('py', doc='Direct python solver interfaces')
class pywrapper(OptSolver):
    """Direct python solver interface"""

    def __new__(cls, *args, **kwds):
        mode = kwds.get('solver_io', 'python')
        if mode is None:
            mode = 'python'
        if mode != 'python':
            logging.getLogger('pyomo.solvers').error(
                "Cannot specify IO mode '%s' for direct python solver interface" % mode
            )
            return None
        #
        if not 'solver' in kwds:
            logging.getLogger('pyomo.solvers').warning(
                "No solver specified for direct python solver interface"
            )
            return None
        kwds['solver_io'] = 'python'
        solver = kwds['solver']
        del kwds['solver']
        return SolverFactory(solver, **kwds)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
from pyomo.common.fileutils import find_library, this_file_dir
import os
from pyomo.common.envvar import PYOMO_CONFIG_DIR
import sys
from pyomo.contrib.appsi.cmodel import cmodel_available


class TestCmodelImport(unittest.TestCase):
    def test_import(self):
        pyomo_config_dir = os.path.join(
            PYOMO_CONFIG_DIR,
            "lib",
            "python%s.%s" % sys.version_info[:2],
            "site-packages",
        )
        cmodel_dir = this_file_dir()
        cmodel_dir = os.path.join(cmodel_dir, os.pardir)
        lib = find_library("appsi_cmodel.*", pathlist=pyomo_config_dir)
        if lib is None:
            lib = find_library("appsi_cmodel.*", pathlist=cmodel_dir)
        if lib is not None:
            self.assertTrue(cmodel_available)
        else:
            raise unittest.SkipTest('appsi library file not found')



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import enum
import sys
from pyomo.common import enums

if sys.version_info[:2] >= (3, 11):
    strictEnum = {'boundary': enum.STRICT}
else:
    strictEnum = {}


class TraversalStrategy(enum.Enum, **strictEnum):
    BreadthFirstSearch = 1
    PrefixDepthFirstSearch = 2
    PostfixDepthFirstSearch = 3
    # aliases
    BFS = BreadthFirstSearch
    ParentLastDepthFirstSearch = PostfixDepthFirstSearch
    PostfixDFS = PostfixDepthFirstSearch
    ParentFirstDepthFirstSearch = PrefixDepthFirstSearch
    PrefixDFS = PrefixDepthFirstSearch
    DepthFirstSearch = PrefixDepthFirstSearch
    DFS = DepthFirstSearch


class SortComponents(enum.Flag, **strictEnum):
    """
    This class is a convenient wrapper for specifying various sort
    ordering.  We pass these objects to the "sort" argument to various
    accessors / iterators to control how much work we perform sorting
    the resultant list.  The idea is that
    "sort=SortComponents.deterministic" is more descriptive than
    "sort=True".
    """

    UNSORTED = 0
    # Note: skip '1' so that we can map True to something other than 1
    ORDERED_INDICES = 2
    SORTED_INDICES = 4
    ALPHABETICAL = 8

    # aliases
    # TODO: deprecate some of these
    unsorted = UNSORTED
    indices = SORTED_INDICES
    declOrder = UNSORTED
    declarationOrder = declOrder
    alphaOrder = ALPHABETICAL
    alphabeticalOrder = alphaOrder
    alphabetical = alphaOrder
    # both alpha and decl orders are deterministic, so only must sort indices
    deterministic = ORDERED_INDICES
    sortBoth = indices | alphabeticalOrder  # Same as True
    alphabetizeComponentAndIndex = sortBoth

    @classmethod
    def _missing_(cls, value):
        if type(value) is bool:
            if value:
                return cls.SORTED_INDICES | cls.ALPHABETICAL
            else:
                return cls.UNSORTED
        elif value is None:
            return cls.UNSORTED
        return super()._missing_(value)

    @staticmethod
    def default():
        return SortComponents.UNSORTED

    @staticmethod
    def sorter(sort_by_names=False, sort_by_keys=False):
        sort = SortComponents.default()
        if sort_by_names:
            sort |= SortComponents.ALPHABETICAL
        if sort_by_keys:
            sort |= SortComponents.SORTED_INDICES
        return sort

    @staticmethod
    def sort_names(flag):
        return SortComponents.ALPHABETICAL in SortComponents(flag)

    @staticmethod
    def sort_indices(flag):
        return SortComponents.SORTED_INDICES in SortComponents(flag)


class VarCollector(enums.IntEnum):
    FromVarComponents = 1
    FromExpressions = 2



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Transformation to propagate state through an equality set."""

from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.core.base.constraint import Constraint
from pyomo.core.base.transformation import TransformationFactory
from pyomo.core.base.suffix import Suffix
from pyomo.core.expr.numvalue import value
from pyomo.core.plugins.transform.hierarchy import IsomorphicTransformation
from pyomo.repn.standard_repn import generate_standard_repn
from pyomo.common.config import (
    ConfigBlock,
    ConfigValue,
    document_kwargs_from_configdict,
)
from pyomo.common.errors import InfeasibleConstraintException


def _build_equality_set(m):
    """Construct an equality set map.

    Maps all variables to the set of variables that are linked to them by
    equality. Mapping takes place using id(). That is, if you have x = y, then
    you would have id(x) -> ComponentSet([x, y]) and id(y) -> ComponentSet([x,
    y]) in the mapping.

    """
    #: dict: map of var UID to the set of all equality-linked var UIDs
    eq_var_map = ComponentMap()
    relevant_vars = ComponentSet()
    for constr in m.component_data_objects(
        ctype=Constraint, active=True, descend_into=True
    ):
        # Check to make sure the constraint is of form v1 - v2 == 0
        if (
            value(constr.lower) == 0
            and value(constr.upper) == 0
            and constr.body.polynomial_degree() == 1
        ):
            repn = generate_standard_repn(constr.body)
            # only take the variables with nonzero coefficients
            vars_ = [v for i, v in enumerate(repn.linear_vars) if repn.linear_coefs[i]]
            if (
                len(vars_) == 2
                and repn.constant == 0
                and sorted(l for l in repn.linear_coefs if l) == [-1, 1]
            ):
                # this is an a == b constraint.
                v1 = vars_[0]
                v2 = vars_[1]
                set1 = eq_var_map.get(v1, ComponentSet([v1]))
                set2 = eq_var_map.get(v2, ComponentSet([v2]))
                relevant_vars.update([v1, v2])
                set1.update(set2)  # set1 is now the union
                for v in set1:
                    eq_var_map[v] = set1

    return eq_var_map, relevant_vars


def _detect_fixed_variables(m):
    """Detect fixed variables due to constraints of form var = const."""
    new_fixed_vars = ComponentSet()
    for constr in m.component_data_objects(
        ctype=Constraint, active=True, descend_into=True
    ):
        if constr.equality and constr.body.polynomial_degree() == 1:
            repn = generate_standard_repn(constr.body)
            if len(repn.linear_vars) == 1 and repn.linear_coefs[0]:
                var = repn.linear_vars[0]
                coef = float(repn.linear_coefs[0])
                const = repn.constant
                var_val = (value(constr.lower) - value(const)) / coef
                var.fix(var_val)
                new_fixed_vars.add(var)
    return new_fixed_vars


@TransformationFactory.register(
    'contrib.propagate_fixed_vars',
    doc="Propagate variable fixing for equalities of type x = y.",
)
@document_kwargs_from_configdict('CONFIG')
class FixedVarPropagator(IsomorphicTransformation):
    """Propagate variable fixing for equalities of type :math:`x = y`.

    If :math:`x` is fixed and :math:`y` is not fixed, then this transformation
    will fix :math:`y` to the value of :math:`x`.

    This transformation can also be performed as a temporary transformation,
    whereby the transformed variables are saved and can be later unfixed.

    Keyword arguments below are specified for the ``apply_to`` and
    ``create_using`` functions.

    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "tmp",
        ConfigValue(
            default=False,
            domain=bool,
            description="True to store the set of transformed variables and "
            "their old states so that they can be later restored.",
        ),
    )

    def _apply_to(self, instance, **kwds):
        config = self.CONFIG(kwds)
        if config.tmp and not hasattr(instance, '_tmp_propagate_fixed'):
            instance._tmp_propagate_fixed = ComponentSet()
        eq_var_map, relevant_vars = _build_equality_set(instance)
        #: ComponentSet: The set of all fixed variables
        fixed_vars = ComponentSet((v for v in relevant_vars if v.fixed))
        newly_fixed = _detect_fixed_variables(instance)
        if config.tmp:
            instance._tmp_propagate_fixed.update(newly_fixed)
        fixed_vars.update(newly_fixed)
        processed = ComponentSet()
        # Go through each fixed variable to propagate the 'fixed' status to all
        # equality-linked variables.
        for v1 in fixed_vars:
            # If we have already processed the variable, skip it.
            if v1 in processed:
                continue

            eq_set = eq_var_map.get(v1, ComponentSet([v1]))
            for v2 in eq_set:
                if v2.fixed and value(v1) != value(v2):
                    raise InfeasibleConstraintException(
                        'Variables {} and {} have conflicting fixed '
                        'values of {} and {}, but are linked by '
                        'equality constraints.'.format(
                            v1.name, v2.name, value(v1), value(v2)
                        )
                    )
                elif not v2.fixed:
                    v2.fix(value(v1))
                    if config.tmp:
                        instance._tmp_propagate_fixed.add(v2)
            # Add all variables in the equality set to the set of processed
            # variables.
            processed.update(eq_set)

    def revert(self, instance):
        """Revert variables fixed by the transformation."""
        for var in instance._tmp_propagate_fixed:
            var.unfix()
        del instance._tmp_propagate_fixed


@TransformationFactory.register(
    'contrib.propagate_eq_var_bounds',
    doc="Propagate variable bounds for equalities of type x = y.",
)
@document_kwargs_from_configdict('CONFIG')
class VarBoundPropagator(IsomorphicTransformation):
    """Propagate variable bounds for equalities of type :math:`x = y`.

    If :math:`x` has a tighter bound then :math:`y`, then this transformation
    will adjust the bounds on :math:`y` to match those of :math:`x`.

    Keyword arguments below are specified for the ``apply_to`` and
    ``create_using`` functions.

    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "tmp",
        ConfigValue(
            default=False,
            domain=bool,
            description="True to store the set of transformed variables and "
            "their old states so that they can be later restored.",
        ),
    )

    def _apply_to(self, instance, **kwds):
        config = self.CONFIG(kwds)
        if config.tmp and not hasattr(instance, '_tmp_propagate_original_bounds'):
            instance._tmp_propagate_original_bounds = Suffix(direction=Suffix.LOCAL)
        eq_var_map, relevant_vars = _build_equality_set(instance)
        processed = ComponentSet()
        # Go through each variable in an equality set to propagate the variable
        # bounds to all equality-linked variables.
        for var in relevant_vars:
            # If we have already processed the variable, skip it.
            if var in processed:
                continue

            var_equality_set = eq_var_map.get(var, ComponentSet([var]))

            #: variable lower bounds in the equality set
            lbs = [v.lb for v in var_equality_set if v.has_lb()]
            max_lb = max(lbs) if len(lbs) > 0 else None
            #: variable upper bounds in the equality set
            ubs = [v.ub for v in var_equality_set if v.has_ub()]
            min_ub = min(ubs) if len(ubs) > 0 else None

            # Check for error due to bound cross-over
            if max_lb is not None and min_ub is not None and max_lb > min_ub:
                # the lower bound is above the upper bound. Raise an exception.
                # get variable with the highest lower bound
                v1 = next(v for v in var_equality_set if v.lb == max_lb)
                # get variable with the lowest upper bound
                v2 = next(v for v in var_equality_set if v.ub == min_ub)
                raise InfeasibleConstraintException(
                    'Variable {} has a lower bound {} '
                    '> the upper bound {} of variable {}, '
                    'but they are linked by equality constraints.'.format(
                        v1.name, value(v1.lb), value(v2.ub), v2.name
                    )
                )

            for v in var_equality_set:
                if config.tmp:
                    # TODO warn if overwriting
                    instance._tmp_propagate_original_bounds[v] = (v.lb, v.ub)
                v.setlb(max_lb)
                v.setub(min_ub)

            processed.update(var_equality_set)

    def revert(self, instance):
        """Revert variable bounds."""
        for v in instance._tmp_propagate_original_bounds:
            old_LB, old_UB = instance._tmp_propagate_original_bounds[v]
            v.setlb(old_LB)
            v.setub(old_UB)
        del instance._tmp_propagate_original_bounds



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections
import functools
import inspect
import itertools

from collections.abc import Sequence
from collections.abc import Mapping

from pyomo.common.autoslots import AutoSlots
from pyomo.common.dependencies import numpy, numpy_available, pandas, pandas_available
from pyomo.common.modeling import NOTSET
from pyomo.core.pyomoobject import PyomoObject

initializer_map = {}
sequence_types = set()
# initialize with function, method, and method-wrapper types.
function_types = set(
    [
        type(PyomoObject.is_expression_type),
        type(PyomoObject().is_expression_type),
        type(PyomoObject.is_expression_type.__call__),
    ]
)


def Initializer(
    arg,
    allow_generators=False,
    treat_sequences_as_mappings=True,
    arg_not_specified=None,
    additional_args=0,
):
    """Standardized processing of Component keyword arguments

    Component keyword arguments accept a number of possible inputs, from
    scalars to dictionaries, to functions (rules) and generators.  This
    function standardizes the processing of keyword arguments and
    returns "initializer classes" that are specialized to the specific
    data type provided.

    Parameters
    ----------
    arg:

        The argument passed to the component constructor.  This could
        be almost any type, including a scalar, dict, list, function,
        generator, or None.

    allow_generators: bool

        If False, then we will raise an exception if ``arg`` is a generator

    treat_sequences_as_mappings: bool

        If True, then if ``arg`` is a sequence, we will treat it as if
        it were a mapping (i.e., ``dict(enumerate(arg))``).  Otherwise
        sequences will be returned back as the value of the initializer.

    arg_not_specified:

        If ``arg`` is ``arg_not_specified``, then the function will
        return None (and not an InitializerBase object).

    additional_args: int

        The number of additional arguments that will be passed to any
        function calls (provided *before* the index value).

    """
    if arg is arg_not_specified:
        return None
    if additional_args:
        if arg.__class__ in function_types:
            if allow_generators or inspect.isgeneratorfunction(arg):
                raise ValueError(
                    "Generator functions are not allowed when passing additional args"
                )
            _args = inspect.getfullargspec(arg)
            _nargs = len(_args.args)
            if inspect.ismethod(arg) and arg.__self__ is not None:
                # Ignore 'self' for bound instance methods and 'cls' for
                # @classmethods
                _nargs -= 1
            if _nargs == 1 + additional_args and _args.varargs is None:
                return ParameterizedScalarCallInitializer(arg, constant=True)
            else:
                return ParameterizedIndexedCallInitializer(arg)
        else:
            base_initializer = Initializer(
                arg=arg,
                allow_generators=allow_generators,
                treat_sequences_as_mappings=treat_sequences_as_mappings,
                arg_not_specified=arg_not_specified,
            )
            if type(base_initializer) in (
                ScalarCallInitializer,
                IndexedCallInitializer,
            ):
                # This is an edge case: if we are providing additional
                # args, but this is the first time we are seeing a
                # callable type, we will (potentially) incorrectly
                # categorize this as an IndexedCallInitializer.  Re-try
                # now that we know this is a function_type.
                return Initializer(
                    arg=base_initializer._fcn,
                    allow_generators=allow_generators,
                    treat_sequences_as_mappings=treat_sequences_as_mappings,
                    arg_not_specified=arg_not_specified,
                    additional_args=additional_args,
                )
            return ParameterizedInitializer(base_initializer)
    if arg.__class__ in initializer_map:
        return initializer_map[arg.__class__](arg)
    if arg.__class__ in sequence_types:
        if treat_sequences_as_mappings:
            return ItemInitializer(arg)
        else:
            return ConstantInitializer(arg)
    if arg.__class__ in function_types:
        # Note: we do not use "inspect.isfunction or inspect.ismethod"
        # because some function-like things (notably cythonized
        # functions) return False
        if not allow_generators and inspect.isgeneratorfunction(arg):
            raise ValueError("Generator functions are not allowed")
        # Historically pyomo.core.base.misc.apply_indexed_rule
        # accepted rules that took only the parent block (even for
        # indexed components).  We will preserve that functionality
        # here.
        #
        # I was concerned that some builtins aren't compatible with
        # getfullargspec (and would need the same try-except logic as in
        # the partial handling), but I have been unable to come up with
        # an example.  The closest was getattr(), but that falls back on
        # getattr.__call__, which does support getfullargspec.
        _args = inspect.getfullargspec(arg)
        _nargs = len(_args.args)
        if inspect.ismethod(arg) and arg.__self__ is not None:
            # Ignore 'self' for bound instance methods and 'cls' for
            # @classmethods
            _nargs -= 1
        if _nargs == 1 and _args.varargs is None:
            return ScalarCallInitializer(
                arg, constant=not inspect.isgeneratorfunction(arg)
            )
        else:
            return IndexedCallInitializer(arg)
    if hasattr(arg, '__len__'):
        if isinstance(arg, Mapping):
            initializer_map[arg.__class__] = ItemInitializer
        elif isinstance(arg, Sequence) and not isinstance(arg, str):
            sequence_types.add(arg.__class__)
        elif isinstance(arg, PyomoObject):
            # TODO: Should IndexedComponent inherit from
            # collections.abc.Mapping?
            if arg.is_component_type() and arg.is_indexed():
                initializer_map[arg.__class__] = ItemInitializer
            else:
                initializer_map[arg.__class__] = ConstantInitializer
        elif any(c.__name__ == 'ndarray' for c in arg.__class__.__mro__):
            if numpy_available and isinstance(arg, numpy.ndarray):
                sequence_types.add(arg.__class__)
        elif any(c.__name__ == 'Series' for c in arg.__class__.__mro__):
            if pandas_available and isinstance(arg, pandas.Series):
                sequence_types.add(arg.__class__)
        elif any(c.__name__ == 'DataFrame' for c in arg.__class__.__mro__):
            if pandas_available and isinstance(arg, pandas.DataFrame):
                initializer_map[arg.__class__] = DataFrameInitializer
        else:
            # Note: this picks up (among other things) all string instances
            initializer_map[arg.__class__] = ConstantInitializer
        # recursively call Initializer to pick up the new registration
        return Initializer(
            arg,
            allow_generators=allow_generators,
            treat_sequences_as_mappings=treat_sequences_as_mappings,
            arg_not_specified=arg_not_specified,
        )
    if inspect.isgenerator(arg) or hasattr(arg, 'next') or hasattr(arg, '__next__'):
        # This catches generators and iterators (like enumerate()), but
        # skips "reusable" iterators like range() as well as Pyomo
        # (finite) Set objects [they were both caught by the
        # "hasattr('__len__')" above]
        if not allow_generators:
            raise ValueError("Generators are not allowed")
        # Deepcopying generators is problematic (e.g., it generates a
        # segfault in pypy3 7.3.0).  We will immediately expand the
        # generator into a tuple and then store it as a constant.
        return ConstantInitializer(tuple(arg))
    if type(arg) is functools.partial:
        try:
            _args = inspect.getfullargspec(arg.func)
        except:
            # Inspect doesn't work for some built-in callables (notably
            # 'int').  We will just have to assume this is a "normal"
            # IndexedCallInitializer
            return IndexedCallInitializer(arg)
        _positional_args = set(_args.args)
        for key in arg.keywords:
            _positional_args.discard(key)
        if len(_positional_args) - len(arg.args) == 1 and _args.varargs is None:
            return ScalarCallInitializer(arg)
        else:
            return IndexedCallInitializer(arg)
    if isinstance(arg, InitializerBase):
        return arg
    if isinstance(arg, PyomoObject):
        # We re-check for PyomoObject here, as that picks up / caches
        # non-components like component data objects and expressions
        initializer_map[arg.__class__] = ConstantInitializer
        return ConstantInitializer(arg)
    if callable(arg) and not isinstance(arg, type):
        # We assume any callable thing could be a functor; but, we must
        # filter out types, as we use types as special identifiers that
        # should not be called (e.g., UnknownSetDimen)
        if inspect.isfunction(arg) or inspect.ismethod(arg):
            # Add this to the set of known function types and try again
            function_types.add(type(arg))
        else:
            # Try again, but use the __call__ method (for supporting
            # things like functors and cythonized functions).  __call__
            # is almost certainly going to be a method-wrapper
            arg = arg.__call__
        return Initializer(
            arg,
            allow_generators=allow_generators,
            treat_sequences_as_mappings=treat_sequences_as_mappings,
            arg_not_specified=arg_not_specified,
        )
    initializer_map[arg.__class__] = ConstantInitializer
    return ConstantInitializer(arg)


class InitializerBase(AutoSlots.Mixin, object):
    """Base class for all Initializer objects"""

    __slots__ = ()

    verified = False

    def constant(self):
        """Return True if this initializer is constant across all indices"""
        return False

    def contains_indices(self):
        """Return True if this initializer contains embedded indices"""
        return False

    def indices(self):
        """Return a generator over the embedded indices

        This will raise a RuntimeError if this initializer does not
        contain embedded indices
        """
        raise RuntimeError(
            "Initializer %s does not contain embedded indices" % (type(self).__name__,)
        )


class ConstantInitializer(InitializerBase):
    """Initializer for constant values"""

    __slots__ = ('val', 'verified')

    def __init__(self, val):
        self.val = val
        self.verified = False

    def __call__(self, parent, idx):
        return self.val

    def constant(self):
        return True


class ItemInitializer(InitializerBase):
    """Initializer for dict-like values supporting __getitem__()"""

    __slots__ = ('_dict',)

    def __init__(self, _dict):
        self._dict = _dict

    def __call__(self, parent, idx):
        return self._dict[idx]

    def contains_indices(self):
        return True

    def indices(self):
        try:
            return self._dict.keys()
        except AttributeError:
            return range(len(self._dict))


class DataFrameInitializer(InitializerBase):
    """Initializer for pandas DataFrame values"""

    __slots__ = ('_df', '_column')

    def __init__(self, dataframe, column=None):
        self._df = dataframe
        if column is not None:
            self._column = column
        elif len(dataframe.columns) == 1:
            self._column = dataframe.columns[0]
        else:
            self._column = None

    def __call__(self, parent, idx):
        if self._column is None:
            return self._df.at[idx]
        return self._df.at[idx, self._column]

    def contains_indices(self):
        return True

    def indices(self):
        if self._column is None:
            return itertools.product(self._df.index, self._df)
        return self._df.index


class IndexedCallInitializer(InitializerBase):
    """Initializer for functions and callable objects"""

    __slots__ = ('_fcn',)

    def __init__(self, _fcn):
        self._fcn = _fcn

    def __call__(self, parent, idx, **kwargs):
        # Note: this is called by a component using data from a Set (so
        # any tuple-like type should have already been checked and
        # converted to a tuple; or flattening is turned off and it is
        # the user's responsibility to sort things out.
        if idx.__class__ is tuple:
            return self._fcn(parent, *idx, **kwargs)
        else:
            return self._fcn(parent, idx, **kwargs)


class ParameterizedIndexedCallInitializer(IndexedCallInitializer):
    """IndexedCallInitializer that accepts additional arguments"""

    __slots__ = ()

    def __call__(self, parent, idx, *args, **kwargs):
        if idx.__class__ is tuple:
            return self._fcn(parent, *args, *idx, **kwargs)
        else:
            return self._fcn(parent, *args, idx, **kwargs)


class CountedCallGenerator:
    """Generator implementing the "counted call" initialization scheme

    This generator implements the older "counted call" scheme, where the
    first argument past the parent block is a monotonically-increasing
    integer beginning at `start_at`.
    """

    def __init__(self, ctype, fcn, scalar, parent, idx, start_at):
        # Note: this is called by a component using data from a Set (so
        # any tuple-like type should have already been checked and
        # converted to a tuple; or flattening is turned off and it is
        # the user's responsibility to sort things out.
        self._count = start_at - 1
        if scalar:
            self._fcn = lambda c: self._filter(ctype, fcn(parent, c))
        elif idx.__class__ is tuple:
            self._fcn = lambda c: self._filter(ctype, fcn(parent, c, *idx))
        else:
            self._fcn = lambda c: self._filter(ctype, fcn(parent, c, idx))

    def __iter__(self):
        return self

    def __next__(self):
        self._count += 1
        return self._fcn(self._count)

    next = __next__

    @staticmethod
    def _filter(ctype, x):
        if x is None:
            raise ValueError(
                """Counted %s rule returned None instead of %s.End.
    Counted %s rules of the form fcn(model, count, *idx) will be called
    repeatedly with an increasing count parameter until the rule returns
    %s.End.  None is not a valid return value in this case due to the
    likelihood that an error in the rule can incorrectly return None."""
                % ((ctype.__name__,) * 4)
            )
        return x


class CountedCallInitializer(InitializerBase):
    """Initializer for functions implementing the "counted call" API."""

    # Pyomo has a historical feature for some rules, where the number of
    # times[*1] the rule was called could be passed as an additional
    # argument between the block and the index.  This was primarily
    # supported by Set and ConstraintList.  There were many issues with
    # the syntax, including inconsistent support for jagged (dimen=None)
    # indexing sets, inconsistent support for *args rules, and a likely
    # infinite loop if the rule returned Constraint.Skip.
    #
    # As a slight departure from previous implementations, we will ONLY
    # allow the counted rule syntax when the rule does NOT use *args
    #
    # [*1] The extra argument was one-based, and was only incremented
    #     when a valid value was returned by the rule and added to the
    #     _data.  This was fragile, as returning something like
    #     {Component}.Skip could result in an infinite loop.  This
    #     implementation deviates from that behavior and increments the
    #     counter every time the rule is called.
    #
    # [JDS 6/2019] We will support a slightly restricted but more
    # consistent form of the original implementation for backwards
    # compatibility, but I believe that we should deprecate this syntax
    # entirely.
    __slots__ = ('_fcn', '_is_counted_rule', '_scalar', '_ctype', '_start')

    def __init__(self, obj, _indexed_init, starting_index=1):
        self._fcn = _indexed_init._fcn
        self._is_counted_rule = None
        self._scalar = not obj.is_indexed()
        self._ctype = obj.ctype
        self._start = starting_index
        if self._scalar:
            self._is_counted_rule = True

    def __call__(self, parent, idx):
        # Note: this is called by a component using data from a Set (so
        # any tuple-like type should have already been checked and
        # converted to a tuple; or flattening is turned off and it is
        # the user's responsibility to sort things out.
        if self._is_counted_rule == False:
            if idx.__class__ is tuple:
                return self._fcn(parent, *idx)
            else:
                return self._fcn(parent, idx)
        if self._is_counted_rule == True:
            return CountedCallGenerator(
                self._ctype, self._fcn, self._scalar, parent, idx, self._start
            )

        # Note that this code will only be called once, and only if
        # the object is not a scalar.
        _args = inspect.getfullargspec(self._fcn)
        _nargs = len(_args.args)
        if inspect.ismethod(self._fcn) and self._fcn.__self__ is not None:
            _nargs -= 1
        _len = len(idx) if idx.__class__ is tuple else 1
        if _len + 2 == _nargs:
            self._is_counted_rule = True
        else:
            self._is_counted_rule = False
        return self.__call__(parent, idx)


class ScalarCallInitializer(InitializerBase):
    """Initializer for functions taking only the parent block argument."""

    __slots__ = ('_fcn', '_constant')

    def __init__(self, _fcn, constant=True):
        self._fcn = _fcn
        self._constant = constant

    def __call__(self, parent, idx, **kwargs):
        return self._fcn(parent, **kwargs)

    def constant(self):
        """Return True if this initializer is constant across all indices"""
        return self._constant


class ParameterizedScalarCallInitializer(ScalarCallInitializer):
    """ScalarCallInitializer that accepts additional arguments"""

    __slots__ = ()

    def __call__(self, parent, idx, *args, **kwargs):
        return self._fcn(parent, *args, **kwargs)


class DefaultInitializer(InitializerBase):
    """Initializer wrapper that maps exceptions to default values.


    Parameters
    ----------
    initializer: :py:class`InitializerBase`
        the Initializer instance to wrap

    default:
        the value to return inlieu of the caught exception(s)

    exceptions: Exception or tuple
        the single Exception or tuple of Exceptions to catch and return
        the default value.

    """

    __slots__ = ('_initializer', '_default', '_exceptions')

    def __init__(self, initializer, default, exceptions):
        self._initializer = initializer
        self._default = default
        self._exceptions = exceptions

    def __call__(self, parent, index, **kwargs):
        try:
            return self._initializer(parent, index, **kwargs)
        except self._exceptions:
            return self._default

    def constant(self):
        """Return True if this initializer is constant across all indices"""
        return self._initializer.constant()

    def contains_indices(self):
        """Return True if this initializer contains embedded indices"""
        return self._initializer.contains_indices()

    def indices(self):
        return self._initializer.indices()


class ParameterizedInitializer(InitializerBase):
    """Wrapper to provide additional positional arguments to Initializer objects"""

    __slots__ = ('_base_initializer',)

    def __init__(self, base):
        self._base_initializer = base

    def constant(self):
        """Return True if this initializer is constant across all indices"""
        return self._base_initializer.constant()

    def contains_indices(self):
        """Return True if this initializer contains embedded indices"""
        return self._base_initializer.contains_indices()

    def indices(self):
        """Return a generator over the embedded indices

        This will raise a RuntimeError if this initializer does not
        contain embedded indices
        """
        return self._base_initializer.indices()

    def __call__(self, parent, idx, *args, **kwargs):
        return self._base_initializer(parent, idx)(parent, *args, **kwargs)


class PartialInitializer(InitializerBase):
    """Partial wrapper of an InitializerBase that supplies additional arguments"""

    __slots__ = ('_fcn',)

    def __init__(self, _fcn, *args, **kwargs):
        self._fcn = functools.partial(_fcn, *args, **kwargs)

    def constant(self):
        return self._fcn.func.constant()

    def contains_indices(self):
        return self._fcn.func.contains_indices()

    def indices(self):
        return self._fcn.func.indices()

    def __call__(self, parent, idx, *args, **kwargs):
        # Note that the Initializer.__call__ API is different from the
        # rule API.  As a result, we cannot just inherit from
        # IndexedCallInitializer and must instead implement our own
        # __call__ here.
        return self._fcn(parent, idx, *args, **kwargs)


_bound_sequence_types = collections.defaultdict(None.__class__)


class BoundInitializer(InitializerBase):
    """Initializer wrapper for processing bounds (mapping scalars to 2-tuples)

    Note that this class is meant to mimic the behavior of
    :py:func:`Initializer` and will return ``None`` if the initializer
    that it is wrapping is ``None``.

    Parameters
    ----------
    arg:

        As with :py:func:`Initializer`, this is the raw argument passed
        to the component constructor.

    obj: :py:class:`Component`

        The component that "owns" the initializer.  This initializer
        will treat sequences as mappings only if the owning component is
        indexed and the sequence passed to the initializer is not of
        length 2

    """

    __slots__ = ('_initializer',)

    def __new__(cls, arg=None, obj=NOTSET):
        # The Initializer() function returns None if the initializer is
        # None.  We will mock that behavior by commandeering __new__()
        if arg is None and obj is not NOTSET:
            return None
        else:
            return super().__new__(cls)

    def __init__(self, arg, obj=NOTSET):
        if obj is NOTSET or obj.is_indexed():
            treat_sequences_as_mappings = not (
                isinstance(arg, Sequence)
                and len(arg) == 2
                and not isinstance(arg[0], Sequence)
            )
        else:
            treat_sequences_as_mappings = False
        self._initializer = Initializer(
            arg, treat_sequences_as_mappings=treat_sequences_as_mappings
        )

    def __call__(self, parent, index, **kwargs):
        val = self._initializer(parent, index, **kwargs)
        if _bound_sequence_types[val.__class__]:
            return val
        if _bound_sequence_types[val.__class__] is None:
            _bound_sequence_types[val.__class__] = isinstance(
                val, Sequence
            ) and not isinstance(val, str)
            if _bound_sequence_types[val.__class__]:
                return val
        return (val, val)

    def constant(self):
        """Return True if this initializer is constant across all indices"""
        return self._initializer.constant()

    def contains_indices(self):
        """Return True if this initializer contains embedded indices"""
        return self._initializer.contains_indices()

    def indices(self):
        return self._initializer.indices()
